from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from pydantic import Field
from pymatgen.analysis.chemenv.coordination_environments.chemenv_strategies import (
    SimplestChemenvStrategy,
)
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometries import (
    AllCoordinationGeometries,
)
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometry_finder import (
    LocalGeometryFinder,
)
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import (
    LightStructureEnvironments,
)
from pymatgen.analysis.structure_analyzer import SpacegroupAnalyzer
from pymatgen.core.structure import Molecule, Structure

from emmet.core.material_property import PropertyDoc

if TYPE_CHECKING:
    from emmet.core.types.typing import IdentifierType

from emmet.core.types.pymatgen_types.structure_adapter import (
    MoleculeType,
    StructureType,
)

DEFAULT_DISTANCE_CUTOFF = 1.4
DEFAULT_ANGLE_CUTOFF = 0.3

AVAILABLE_METHODS = {
    "DefaultSimplestChemenvStrategy": SimplestChemenvStrategy(),
    # "DefaultSimplestChemenvStrategy_all_bonds": SimplestChemenvStrategy(additional_condition=0),
}


DEFAULTSIMPLESTCHEMENVSTRATEGY = "Simplest ChemenvStrategy using fixed angle and distance parameters for the definition of neighbors in the Voronoi approach. The coordination environment is then given as the one with the lowest continuous symmetry measure. Options: distance_cutoff=1.4 angle_cutoff=0.3 additional_condition=1 continuous_symmetry_measure_cutoff=10.0"  # noqa: E501
# SIMPLESTCHEMENVSTRATEGY_ALL_BONDS = "Simplest ChemenvStrategy using fixed angle and distance parameters for the definition of neighbors in the Voronoi approach. The coordination environment is then given as the one with the lowest continuous symmetry measure. Options: distance_cutoff=1.4 angle_cutoff=0.3 additional_condition=0 continuous_symmetry_measure_cutoff=10.0"  # noqa: E501


METHODS_DESCRIPTION = {
    "DefaultSimplestChemenvStrategy": DEFAULTSIMPLESTCHEMENVSTRATEGY,
    # "DefaultSimplestChemenvStrategy_all_bonds": SIMPLESTCHEMENVSTRATEGY_ALL_BONDS,
}


COORDINATION_GEOMETRIES = Literal[
    "S:1",
    "L:2",
    "A:2",
    "TL:3",
    "TY:3",
    "TS:3",
    "T:4",
    "S:4",
    "SY:4",
    "SS:4",
    "PP:5",
    "S:5",
    "T:5",
    "O:6",
    "T:6",
    "PP:6",
    "PB:7",
    "ST:7",
    "ET:7",
    "FO:7",
    "C:8",
    "SA:8",
    "SBT:8",
    "TBT:8",
    "DD:8",
    "DDPN:8",
    "HB:8",
    "BO_1:8",
    "BO_2:8",
    "BO_3:8",
    "TC:9",
    "TT_1:9",
    "TT_2:9",
    "TT_3:9",
    "HD:9",
    "TI:9",
    "SMA:9",
    "SS:9",
    "TO_1:9",
    "TO_2:9",
    "TO_3:9",
    "PP:10",
    "PA:10",
    "SBSA:10",
    "MI:10",
    "BS_1:10",
    "BS_2:10",
    "TBSA:10",
    "PCPA:11",
    "H:11",
    "DI:11",
    "I:12",
    "PBP:12",
    "TT:12",
    "C:12",
    "AC:12",
    "SC:12",
    "HP:12",
    "HA:12",
    "SH:13",
    "DD:20",
    "None",
]

COORDINATION_GEOMETRIES_IUPAC = Literal[
    "TOCT-9",
    "CUS-10",
    "PPR-10",
    "PPRP-11",
    "OCF-7",
    "SAPRT-10",
    "T-4",
    "PAPR-10",
    "DD-20",
    "TPR-6",
    "OCT-8",
    "PP-5",
    "SPY-4",
    "HBPY-9",
    "A-2",
    "TPRT-8",
    "TS-3",
    "SS-4",
    "None",
    "SAPR-8",
    "SPY-5",
    "SAPRS-10",
    "HPR-12",
    "L-2",
    "PPY-6",
    "SP-4",
    "IC-12",
    "HBPY-8",
    "TPY-3",
    "CUS-9",
    "TPRS-8",
    "DD-8",
    "TBPY-5",
    "TP-3",
    "TCA-9",
    "PPRP-12",
    "SAPRS-9",
    "OC-6",
    "PBPY-7",
    "TPRS-7",
    "TPRS-9",
    "CU-8",
    "HAPR-12",
    "TPRT-7",
]

COORDINATION_GEOMETRIES_IUCR = Literal[
    "[8do]",
    "[3n]",
    "[6p]",
    "[8cb]",
    "[5by]",
    "[12p]",
    "[6p2c]",
    "[8acb]",
    "[7by]",
    "[3l]",
    "[6p1c]",
    "[12aco]",
    "[2l]",
    "None",
    "[12i]",
    "[6o]",
    "[6p3c]",
    "[4n]",
    "[12co]",
    "[1l]",
    "[5y]",
    "[5l]",
    "[4l]",
    "[2n]",
    "[8by]",
    "[4t]",
    "[12tt]",
]
COORDINATION_GEOMETRIES_NAMES = Literal[
    "Pentagonal pyramid",
    "Metabidiminished icosahedron",
    "Square-face bicapped trigonal prism",
    "Pentagonal plane",
    "Cube",
    "Triangular cupola",
    "Heptagonal dipyramid",
    "Pentagonal prism",
    "T-shaped",
    "Square-face capped trigonal prism",
    "Single neighbor",
    "Triangular-face bicapped trigonal prism",
    "Pentagonal-face capped pentagonal antiprism",
    "Trigonal prism",
    "Square cupola",
    "See-saw",
    "Square non-coplanar",
    "Hexagonal bipyramid",
    "Triangular non-coplanar",
    "Tetrahedron",
    "Tricapped triangular prism (one square-face cap and two triangular-face caps)",
    "Square-face monocapped antiprism",
    "Hexagonal antiprism",
    "Linear",
    "Octahedron",
    "Truncated tetrahedron",
    "Square-face capped hexagonal prism",
    "Tricapped octahedron (all 3 cap faces share one atom)",
    "Bicapped square prism (opposite faces)",
    "Bicapped octahedron (cap faces with one edge in common)",
    "Tridiminished icosahedron",
    "Cuboctahedron",
    "Dodecahedron",
    "Dodecahedron with triangular faces - p2345 plane normalized",
    "Square-face capped square prism",
    "Icosahedron",
    "Dodecahedron with triangular faces",
    "Pentagonal bipyramid",
    "Tricapped octahedron (cap faces are aligned)",
    "Square pyramid",
    "Tricapped triangular prism (three square-face caps)",
    "Trigonal bipyramid",
    "Tricapped triangular prism (two square-face caps and one triangular-face cap)",
    "Pentagonal antiprism",
    "Diminished icosahedron",
    "Anticuboctahedron",
    "Trigonal-face bicapped square antiprism",
    "Bicapped octahedron (cap faces with one atom in common)",
    "Hexagonal prism",
    "Angular",
    "Trigonal plane",
    "Square plane",
    "Bicapped octahedron (opposed cap faces)",
    "Bicapped square prism (adjacent faces)",
    "Square antiprism",
    "Pentagonal-face bicapped pentagonal prism",
    "Square-face bicapped square antiprism",
    "Hendecahedron",
    "End-trigonal-face capped trigonal prism",
    "Face-capped octahedron",
    "Tricapped octahedron (all 3 cap faces are sharingone edge of a face)",
]

COORDINATION_GEOMETRIES_NAMES_WITH_ALTERNATIVES = Literal[
    "Anticuboctahedron (also known as Triangular bicupola)",
    "Trigonal bipyramid (also known as Trigonal dipyramid, Triangular dipyramid)",
    "Octahedron (also known as Square dipyramid, Square bipyramid, Triangular antiprism, Trigonal antiprism)",
    "Diminished icosahedron",
    "Square antiprism (also known as Tetragonal antiprism, Anticube)",
    "Square cupola",
    "Hexagonal bipyramid (also known as Hexagonal dipyramid)",
    "Square-face capped hexagonal prism",
    "Bicapped square prism (opposite faces) (also known as Bicapped cube)",
    "Square-face bicapped trigonal prism",
    "Square non-coplanar",
    "Triangular-face bicapped trigonal prism",
    "Square plane",
    "Trigonal prism (also known as Triangular prism)",
    "Bicapped octahedron (opposed cap faces)",
    "Tricapped octahedron (all 3 cap faces share one atom)",
    "Pentagonal prism",
    "Triangular cupola",
    "Bicapped square prism (adjacent faces) (also known as Bicapped cube)",
    "Cuboctahedron",
    "Tricapped triangular prism (two square-face caps and one triangular-face cap) (also known as Triaugmented trigonal prism)",  # noqa: E501
    "Square-face capped trigonal prism (also known as Augmented triangular prism)",
    "Tetrahedron (also known as Triangular pyramid, Trigonal pyramid)",
    "Cube (also known as Square prism, Tetragonal prism)",
    "Bicapped octahedron (cap faces with one edge in common)",
    "Bicapped octahedron (cap faces with one atom in common)",
    "Pentagonal pyramid",
    "Dodecahedron with triangular faces (also known as Snub disphenoid, Siamese dodecahedron)",
    "Pentagonal antiprism (also known as Paradiminished icosahedron)",
    "Tricapped octahedron (cap faces are aligned)",
    "Icosahedron",
    "Dodecahedron",
    "Face-capped octahedron (also known as Monocapped octahedron)",
    "Angular",
    "Hendecahedron (also known as Bisymmetric hendecahedron)",
    "Trigonal-face bicapped square antiprism",
    "Pentagonal-face capped pentagonal antiprism (also known as Gyroelongated pentagonal pyramid, Diminished icosahedron, Truncated icosahedron)",  # noqa: E501
    "Linear",
    "Pentagonal plane (also known as Pentagon)",
    "Tricapped triangular prism (three square-face caps) (also known as Triaugmented trigonal prism)",
    "Tricapped octahedron (all 3 cap faces are sharingone edge of a face)",
    "Heptagonal dipyramid (also known as Heptagonal bipyramid)",
    "T-shaped",
    "Single neighbor",
    "Trigonal plane (also known as Triangular planar)",
    "Dodecahedron with triangular faces - p2345 plane normalized (also known as Snub disphenoid - p2345 plane normalized, Siamese dodecahedron - p2345 plane normalized)",  # noqa: E501
    "Tricapped triangular prism (one square-face cap and two triangular-face caps) (also known as Triaugmented trigonal prism)",  # noqa: E501
    "Truncated tetrahedron",
    "Hexagonal prism",
    "Tridiminished icosahedron",
    "Metabidiminished icosahedron",
    "See-saw",
    "Square-face capped square prism (also known as Monocapped cube)",
    "Square pyramid",
    "Pentagonal-face bicapped pentagonal prism",
    "Hexagonal antiprism",
    "Triangular non-coplanar",
    "End-trigonal-face capped trigonal prism (also known as Augmented triangular prism)",
    "Pentagonal bipyramid (also known as Pentagonal dipyramid)",
    "Square-face monocapped antiprism (also known as Gyroelongated square pyramid)",
    "Square-face bicapped square antiprism (also known as Square-face bicapped square anticube, Bicapped anticube, Gyroelongated square dipyramid)",  # noqa: E501
]


class ChemEnvDoc(PropertyDoc):
    """
    Coordination environments based on cation-anion bonds computed for all unique cations in this structure.
    If no oxidation states are available, all bonds will be considered as a fall-back.
    """

    property_name: str = "coord_environment"

    structure: StructureType | None = Field(
        ...,
        description="The structure used in the generation of the chemical environment data",
    )

    valences: list[float] = Field(
        description="List of valences for each site in this material to determine cations"
    )

    species: list[str] = Field(
        description="List of unique (cationic) species in structure."
    )

    chemenv_symbol: list[COORDINATION_GEOMETRIES] = Field(
        description="List of ChemEnv symbols for unique (cationic) species in structure"
    )

    chemenv_iupac: list[COORDINATION_GEOMETRIES_IUPAC] = Field(
        description="List of symbols for unique (cationic) species in structure in IUPAC format"
    )

    chemenv_iucr: list[COORDINATION_GEOMETRIES_IUCR] = Field(
        description="List of symbols for unique (cationic) species in structure in IUCR format"
    )

    chemenv_name: list[COORDINATION_GEOMETRIES_NAMES] = Field(
        description="List of text description of coordination environment for unique (cationic) species in structure."
    )
    chemenv_name_with_alternatives: list[
        COORDINATION_GEOMETRIES_NAMES_WITH_ALTERNATIVES
    ] = Field(
        description="List of text description of coordination environment including alternative descriptions for unique (cationic) species in structure."  # noqa: E501
    )

    csm: list[float | None] = Field(
        description="Saves the continous symmetry measures for unique (cationic) species in structure"
    )

    method: str | None = Field(
        description="Method used to compute chemical environments"
    )

    mol_from_site_environments: list[MoleculeType | None] = Field(
        description="List of Molecule Objects describing the detected environment."
    )

    wyckoff_positions: list[str] = Field(
        description="List of Wyckoff positions for unique (cationic) species in structure."
    )

    warnings: str | None = Field(None, description="Warning")

    @classmethod
    def from_structure(
        cls,
        structure: Structure,
        material_id: IdentifierType | None = None,
        **kwargs,
    ):  # type: ignore[override]
        """

        Args:
            structure: structure including oxidation states
            material_id: mpid
            **kwargs:

        Returns:

        """
        d = {
            "valences": [],
            "species": [],
            "chemenv_symbol": [],
            "chemenv_iupac": [],
            "chemenv_iucr": [],
            "chemenv_name": [],
            "chemenv_name_with_alternatives": [],
            "csm": [],
            "mol_from_site_environments": [],
            "wyckoff_positions": [],
            "method": None,
            "warnings": None,
            # "structure_environment": None,
        }  # type: dict

        # try:
        list_mol = []
        list_chemenv = []
        list_chemenv_iupac = []
        list_chemenv_iucr = []
        list_chemenv_text = []
        list_chemenv_text_with_alternatives = []
        list_csm = []
        list_wyckoff = []
        list_species = []

        valences = [getattr(site.specie, "oxi_state", None) for site in structure]
        valences = [v for v in valences if v is not None]
        sga = SpacegroupAnalyzer(structure)
        symm_struct = sga.get_symmetrized_structure()
        # We still need the whole list of indices
        inequivalent_indices = [
            indices[0] for indices in symm_struct.equivalent_indices
        ]
        # if len(inequivalent_indices)>5:
        #    print("Too many sites")
        #    raise Exception
        # wyckoff symbols for all inequivalent indices
        wyckoffs_unique = symm_struct.wyckoff_symbols
        # use the local geometry finder to get the important information
        lgf = LocalGeometryFinder()
        lgf.setup_structure(structure=structure)
        all_ce = AllCoordinationGeometries()
        if len(valences) == len(structure):
            # Standard alorithm will only focus on cations and cation-anion bonds!
            method_description = "DefaultSimplestChemenvStrategy"
            method = AVAILABLE_METHODS[method_description]
            # We will only focus on cations!
            inequivalent_indices_cations = [
                indices[0]
                for indices in symm_struct.equivalent_indices
                if valences[indices[0]] > 0.0  # type: ignore[operator]
            ]

            se = lgf.compute_structure_environments(
                only_indices=inequivalent_indices_cations,
                valences=valences,
                maximum_distance_factor=DEFAULT_DISTANCE_CUTOFF,
                minimum_angle_factor=DEFAULT_ANGLE_CUTOFF,
            )
            lse = LightStructureEnvironments.from_structure_environments(
                strategy=method, structure_environments=se
            )
            warnings = None
        else:
            d.update(
                {
                    "warnings": "No oxidation states available. Cation-anion bonds cannot be identified."
                }
            )
            return super().from_structure(
                meta_structure=structure,
                material_id=material_id,
                **d,
                **kwargs,
            )
            # method_description = "DefaultSimplestChemenvStrategy_all_bonds"
            # method = AVAILABLE_METHODS[method_description]

            # se = lgf.compute_structure_environments(
            #     only_indices=inequivalent_indices,
            # )
            # lse = LightStructureEnvironments.from_structure_environments(strategy=method,
            #     structure_environments=se)
            # Trick to get rid of duplicate code
            # inequivalent_indices_cations = inequivalent_indices
            # warnings = "No oxidation states. Analysis will now include all bonds"

        for index, wyckoff in zip(inequivalent_indices, wyckoffs_unique):
            # ONLY CATIONS
            if index in inequivalent_indices_cations:
                # Coordinaton environment will be saved as a molecule!
                mol = Molecule.from_sites(
                    [structure[index]] + lse.neighbors_sets[index][0].neighb_sites
                )
                mol = mol.get_centered_molecule()
                env = lse.coordination_environments[index]
                co = all_ce.get_geometry_from_mp_symbol(env[0]["ce_symbol"])
                name = co.name
                if co.alternative_names:
                    name += f" (also known as {', '.join(co.alternative_names)})"
                # save everything in a list
                list_chemenv_text.append(co.name)
                list_chemenv_text_with_alternatives.append(name)
                list_chemenv.append(co.ce_symbol)
                list_chemenv_iucr.append(co.IUCr_symbol_str)
                list_chemenv_iupac.append(co.IUPAC_symbol_str)
                list_mol.append(mol)
                list_wyckoff.append(wyckoff)
                list_species.append(structure[index].species_string)
                list_csm.append(env[0]["csm"])

            d.update(
                {
                    "valences": valences,
                    "species": list_species,
                    "chemenv_symbol": list_chemenv,
                    "chemenv_iupac": list_chemenv_iupac,
                    "chemenv_iucr": list_chemenv_iucr,
                    "chemenv_name": list_chemenv_text,
                    "chemenv_name_with_alternatives": list_chemenv_text_with_alternatives,
                    "csm": list_csm,
                    "mol_from_site_environments": list_mol,
                    "wyckoff_positions": list_wyckoff,
                    # "structure_environment": se.as_dict(),
                    "method": METHODS_DESCRIPTION[method_description],
                    "warnings": warnings,
                }
            )

        # except Exception:
        #     d.update({"warnings": "ChemEnv algorithm failed."})

        return super().from_structure(
            meta_structure=structure,
            material_id=material_id,
            **d,
            **kwargs,
        )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.phonon.dos import PhononDos
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.lattice_adapter import MSONableTypedLatticeDict
from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)

TypedPhononBandStructureSymmLineDict = TypedDict(
    "TypedPhononBandStructureSymmLineDict",
    {
        "@module": str,
        "@class": str,
        "bands": list[list[float]],
        "eigendisplacements": dict[str, list[list[list[list[float]]]]],
        "has_nac": bool,
        "labels_dict": dict[str, list[float, float, float]],  # type: ignore[type-arg]
        "lattice_rec": MSONableTypedLatticeDict,
        "qpoints": list[list[float]],
        "structure": TypedStructureDict,
    },
)

TypedPhononDosDict = TypedDict(
    "TypedPhononDosDict",
    {
        "@module": str,
        "@class": str,
        "densities": list[float],
        "frequencies": list[float],
        "pdos": list[list[float]],
        "structure": TypedStructureDict,
    },
)

PhononBandStructureSymmLineTypeVar = TypeVar(
    "PhononBandStructureSymmLineTypeVar",
    PhononBandStructureSymmLine,
    TypedPhononBandStructureSymmLineDict,
)

PhononDosTypeVar = TypeVar("PhononDosTypeVar", PhononDos, TypedPhononDosDict)


def pop_empty_keys_from_structure(
    d: PhononDosTypeVar | PhononBandStructureSymmLineTypeVar,
):
    if isinstance(d, dict):
        d["structure"] = pop_empty_structure_keys(d["structure"])

    return d


AnnotatedPhononBandStructureSymmLine = Annotated[
    PhononBandStructureSymmLineTypeVar, BeforeValidator(pop_empty_keys_from_structure)
]

AnnotatedPhononDos = Annotated[
    PhononDosTypeVar, BeforeValidator(pop_empty_keys_from_structure)
]



import os
from pathlib import Path
import requests
from emmet.core.mpid import AlphaID, VALID_ALPHA_SEPARATORS

GH_OBSCENITY = "https://raw.githubusercontent.com/LDNOOBW/List-of-Dirty-Naughty-Obscene-and-Otherwise-Bad-Words/master/"
LANGUAGES = [
    "ar",
    "cs",
    "da",
    "de",
    "en",
    "eo",
    "es",
    "fa",
    "fi",
    "fil",
    "fr",
    "fr-CA-u-sd-caqc",
    "hi",
    "hu",
    "it",
    "ja",
    "kab",
    "ko",
    "nl",
    "no",
    "pl",
    "pt",
    "ru",
    "sv",
    "th",
    "tlh",
    "tr",
    "zh",
]
DEFAULT_FILE = (
    Path(__file__).parent / ".." / "emmet" / "core" / "_forbidden_alpha_id.py"
)


def generate_exclude_list(
    github_url: str = GH_OBSCENITY,
    languages: list[str] = LANGUAGES,
    output_file: str | Path = DEFAULT_FILE,
) -> list[int]:
    """Generate a list of integers corresponding to multi-lingual obscene words.

    Parameters
    -----------
    github_url: str = GH_OBSCENITY
        The github repo URL to fetch obscene word lists from
    languages: list[str] = LANGUAGES
        The languages in the repo to select word lists for
    output_file: str | Path = DEFAULT_FILE
        The file to write the integer list to

    Returns
    -----------
    list of int
        AlphaID integer values of those words/phrases
    """

    obscene_alpha_ids = set()
    for lang in languages:
        for word in (
            requests.get(f"{os.path.join(github_url,lang)}")
            .content.decode()
            .splitlines()
        ):
            for sep in VALID_ALPHA_SEPARATORS | {" "}:
                word = word.replace(sep, "")

            alpha_id = None
            try:
                alpha_id = AlphaID(word)
            except Exception:
                continue

            # Only retain "obscene AlphaIDs" that
            # are longer than two characters, with leading "a"'s removed
            if len(AlphaID(int(alpha_id))) > 2:
                obscene_alpha_ids.update({alpha_id})

    if any(alpha_id._prefix for alpha_id in obscene_alpha_ids):
        badly_parsed = "\n".join(
            alpha_id._identifier for alpha_id in obscene_alpha_ids if alpha_id._prefix
        )
        raise ValueError(
            "None of the obscene AlphaIDs should have prefixes or separators. "
            "Please check the following strings, which have been parsed incorrectly:\n"
            f"{badly_parsed}"
        )

    obscene_alpha_id = sorted(obscene_alpha_ids)
    output_str = """\"\"\"Define AlphaID integer values which cannot be used.\"\"\"

FORBIDDEN_ALPHA_ID_VALUES : set[int] = {
"""
    output_str += (
        ",\n".join(f"   {int(alpha_id)}" for alpha_id in obscene_alpha_id) + "\n}"
    )

    with open(output_file, "w+") as f:
        f.write(output_str)

    return obscene_alpha_id


if __name__ == "__main__":

    _ = generate_exclude_list()



"""Test NEB document class."""

from pathlib import Path
import pytest
import shutil
from tempfile import TemporaryDirectory

from monty.serialization import loadfn
from pymatgen.core import Structure

from emmet.core.neb import (
    NebTaskDoc,
    NebMethod,
    BarrierAnalysis,
    NebIntermediateImagesDoc,
)
from emmet.core.tasks import InputDoc, OrigInputs
from emmet.core.vasp.calculation import Calculation, CalculationInput

from emmet.core.testing_utils import assert_schemas_equal


@pytest.fixture(scope="module")
def neb_test_dir(test_dir: Path) -> Path:
    return test_dir / "neb"


@pytest.mark.parametrize("from_dir", [True, False])
def test_neb_task_doc(neb_test_dir, from_dir: bool):
    if from_dir:
        with TemporaryDirectory() as tmpdir:
            shutil.unpack_archive(neb_test_dir / "neb_sample_calc.zip", tmpdir, "zip")
            neb_doc = NebIntermediateImagesDoc.from_directory(Path(tmpdir) / "neb")
        num_images = 3
        intermed_idxs = (0, 1, 2)
        assert isinstance(neb_doc.orig_inputs, OrigInputs)
        assert isinstance(neb_doc.inputs, InputDoc)
    else:
        neb_doc = NebTaskDoc(**loadfn(neb_test_dir / "Si_neb_doc.json.bz2"))
        num_images = 5
        intermed_idxs = (1, 2, 3)
        assert neb_doc.num_images == num_images
    neb_doc_dict = neb_doc.model_dump()

    assert len(neb_doc.images) == num_images
    assert len(neb_doc.energies) == num_images
    assert len(neb_doc.images) == num_images

    # test that NEB image calculations are all VASP Calculation objects
    assert len(neb_doc.calculations) == num_images
    for image_idx, image_calc in enumerate(neb_doc.calculations):
        assert_schemas_equal(image_calc, neb_doc_dict["calculations"][image_idx])
        assert isinstance(image_calc, Calculation)

    assert all(isinstance(doc.input, CalculationInput) for doc in neb_doc.calculations)

    # check NEB method parsing
    assert neb_doc.neb_method == NebMethod.CLIMBING_IMAGE

    # check that all structures exist
    assert all(isinstance(image, Structure) for image in neb_doc.images)

    # Check that image calculation dirs all have common root / expected format
    assert all(
        neb_doc.directories[idx].startswith(str(neb_doc.dir_name))
        for idx in intermed_idxs
    )

    assert all(
        Path(neb_doc.directories[idx]) == Path(neb_doc.dir_name) / f"{image_idx+1:02}"
        for image_idx, idx in enumerate(intermed_idxs)
    )

    # Check that VASP objects stored for each image calc
    assert len(neb_doc.objects) == num_images

    assert all(
        energy == pytest.approx(neb_doc.calculations[idx].output.energy)
        for idx, energy in enumerate(neb_doc.energies)
    )


def test_from_directories(neb_test_dir):
    with TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        shutil.unpack_archive(neb_test_dir / "neb_sample_calc.zip", tmpdir, "zip")
        neb_doc = NebTaskDoc.from_directories(
            [tmpdir / f"relax_endpoint_{idx+1}" for idx in range(2)],
            tmpdir / "neb",
        )

    assert all(isinstance(calc, Calculation) for calc in neb_doc.calculations)

    for i, ep_dir in enumerate(neb_doc.directories):
        if i == 0 or i == neb_doc.num_images - 1:
            assert "relax_endpoint_" in ep_dir
        else:
            assert f"neb/{i:02}" in ep_dir

    assert len(neb_doc.energies) == neb_doc.num_images
    assert len(neb_doc.images) == neb_doc.num_images
    assert isinstance(neb_doc.barrier_analysis, BarrierAnalysis)

    assert all(
        getattr(neb_doc.barrier_analysis, k, None) is not None
        for k in (
            "energies",
            "frame_index",
            "cubic_spline_pars",
            "ts_frame_index",
            "ts_energy",
            "ts_in_frames",
            "forward_barrier",
            "reverse_barrier",
        )
    )

    assert all(
        getattr(neb_doc, f"{direction}_barrier")
        == getattr(neb_doc.barrier_analysis, f"{direction}_barrier", None)
        for direction in ("forward", "reverse")
    )



"""Define utility functions used across the emmet namespace packages."""

from __future__ import annotations

import copy
import datetime
import hashlib
import inspect
import logging
from enum import Enum
from importlib import import_module
from itertools import groupby
from math import gcd
from typing import TYPE_CHECKING, get_args

import numpy as np
from monty.json import MontyDecoder, MSONable
from pydantic import BaseModel, RootModel
from pydantic._internal._utils import lenient_issubclass
from pymatgen.analysis.elasticity.strain import Deformation
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.analysis.local_env import OpenBabelNN, metal_edge_extender
from pymatgen.analysis.molecule_matcher import MoleculeMatcher
from pymatgen.analysis.structure_matcher import (
    AbstractComparator,
    ElementComparator,
    StructureMatcher,
)
from pymatgen.core.structure import Molecule, Structure
from pymatgen.transformations.standard_transformations import (
    DeformStructureTransformation,
)
from pymatgen.util.graph_hashing import weisfeiler_lehman_graph_hash
from typing_extensions import TypedDict

from emmet.core import ARROW_COMPATIBLE
from emmet.core.mpid import MPculeID
from emmet.core.settings import EmmetSettings

try:
    import blake3
except ImportError:
    blake3 = None  # type: ignore

try:
    import bson
except ImportError:
    bson = None  # type: ignore


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import Any

    from emmet.core.types.typing import FSPathType

logger = logging.getLogger(__name__)

SETTINGS = EmmetSettings()


def type_override(overrides: dict[str, Any]):
    """
    Pydantic model decorator for declaring a field should be serialized
    as a type that does not match the type hint for the field. Examples:
    narrow a union type to a single value or hint at a the return type
    of a field_serializer used during model dumping.
    """

    def wrapped(cls):
        cls.type_overrides = {**getattr(cls, "type_overrides", {}), **overrides}
        return cls

    return wrapped


def set_msonable_type_adapter(cls):
    """
    Decorator for MSONables defined in emmet with arrow compatible
    type hints. Shortcut for generating type adapters rather than
    writing a dedicated stub file.
    """
    TypedClsDict = TypedDict(
        f"Typed{cls.__name__}Dict",
        {
            "@module": str,
            "@class": str,
            "@version": str,
            **{
                key: field.annotation
                for key, field in inspect.signature(cls).parameters.items()
            },
        },
    )

    class ClsTypeAdapter(RootModel):
        root: TypedClsDict

    setattr(cls, "__type_adapter__", ClsTypeAdapter)

    return cls


def arrow_incompatible(cls):
    """
    Simple decorator to mark a Pydantic model as being incompatible with
    serialization using pyarrow. This should only be applied as a temporary
    measure, all document models should aim for full type introspection.
    """
    cls.arrow_incompatible = True
    return cls


def get_sg(struc, symprec=SETTINGS.SYMPREC) -> int:
    """helper function to get spacegroup with a loose tolerance"""
    try:
        return struc.get_space_group_info(symprec=symprec)[1]
    except Exception:
        return -1


def get_num_formula_units(composition: dict[Any, int | float]) -> int:
    """Get the number of formula units in a dict-like composition.

    This implementation differs slightly from how some pymatgen/atomate2
    internals work. In those, certain formulas, e.g., N, will assume
    a smallest formula unit of N2. Thus even if a specified composition is
    `{"N": 1}`, the reduced composition will be `{"N": 2}`, and the number of
    formula units 1/2.

    This always just returns the greatest common divisor of a composition.
    """
    num_form_u = 1
    if all(abs(int(val) - val) < 1e-6 for val in composition.values()):
        num_form_u = gcd(*[int(sc) for sc in composition.values()])
    return num_form_u


def group_structures(
    structures: list[Structure],
    ltol: float = SETTINGS.LTOL,
    stol: float = SETTINGS.STOL,
    angle_tol: float = SETTINGS.ANGLE_TOL,
    symprec: float = SETTINGS.SYMPREC,
    comparator: AbstractComparator = ElementComparator(),
) -> Iterator[list[Structure]]:
    """
    Groups structures according to space group and structure matching

    Args:
        structures ([Structure]): list of structures to group
        ltol (float): StructureMatcher tuning parameter for matching tasks to materials
        stol (float): StructureMatcher tuning parameter for matching tasks to materials
        angle_tol (float): StructureMatcher tuning parameter for matching tasks to materials
        symprec (float): symmetry tolerance for space group finding
    """

    sm = StructureMatcher(
        ltol=ltol,
        stol=stol,
        angle_tol=angle_tol,
        primitive_cell=True,
        scale=True,
        attempt_supercell=False,
        allow_subset=False,
        comparator=comparator,
    )

    def _get_sg(struc):
        return get_sg(struc, symprec=symprec)

    # First group by spacegroup number then by structure matching
    for _, pregroup in groupby(sorted(structures, key=_get_sg), key=_get_sg):
        for group in sm.group_structures(list(pregroup)):
            yield group


def undeform_structure(structure: Structure, transformations: dict) -> Structure:
    """
    Get an undeformed structure by applying transformations in a reverse order.

    Args:
        structure: deformed structure
        transformation: transformation that deforms the structure

    Returns:
        undeformed structure
    """

    for transformation in reversed(transformations.get("history", [])):
        if transformation["@class"] == "DeformStructureTransformation":
            deform = Deformation(transformation["deformation"])
            dst = DeformStructureTransformation(deform.inv)
            structure = dst.apply_transformation(structure)
        else:
            raise RuntimeError(
                "Expect transformation to be `DeformStructureTransformation`; "
                f"got {transformation['@class']}"
            )

    return structure


def generate_robocrys_condensed_struct_and_description(
    structure: Structure,
    mineral_matcher=None,
    symprecs: list[float] = [0.01, 0.1, 1.0e-3],
) -> tuple[dict[str, Any], Any]:
    """
    Get robocrystallographer description of a structure.

    Input
    ------
    structure : pymatgen .Structure
    mineral_matcher : optional robocrys MineralMatcher object
        Slightly reduces load time by storing mineral data
        in memory, rather than reloading for each structure.
    symprecs : list[float]
        A list of symprec values to try for symmetry identification.
        The first value is the default used by robocrys, then
        the default used by emmet (looser), then a tighter symprec.

    Output
    -------
    A robocrys condensed structure and description.
    """
    try:
        from robocrys import StructureCondenser, StructureDescriber
    except ImportError:
        raise ImportError(
            "robocrys needs to be installed to generate Robocrystallographer descriptions"
        )

    for isymprec, symprec in enumerate(symprecs):
        # occasionally, symmetry detection fails - give a few chances to modify symprec
        try:
            condenser = StructureCondenser(
                mineral_matcher=mineral_matcher, symprec=symprec
            )
            condensed_structure = condenser.condense_structure(structure)
            break
        except ValueError as exc:
            if isymprec == len(symprecs) - 1:
                raise exc

    for desc_fmt in ["unicode", "html", "raw"]:
        try:
            describer = StructureDescriber(
                describe_symmetry_labels=False, fmt=desc_fmt, return_parts=False
            )
            description = describer.describe(condensed_structure)
            break
        except ValueError as exc:
            # pymatgen won't convert a "subscript period" character to unicode
            # in these cases, the description is still generated but unicode
            # parsing failed - use html instead
            if "subscript period" not in str(exc):
                raise exc

    return condensed_structure, description


def group_molecules(molecules: list[Molecule]):
    """
    Groups molecules according to composition, charge, and equality

    Note: this function is (currently) only used in the MoleculesAssociationBuilder.
        At that stage, we want to link calculations that are performed on
        identical structures. Collapsing similar structures on the basis of e.g.
        graph isomorphism happens at a later stage.

    Args:
        molecules (list[Molecule])
    """

    def _mol_form(mol_solv):
        return mol_solv.composition.alphabetical_formula

    # Extremely tight tolerance is desirable
    # We want to match only calculations that are EXACTLY the same
    # Molecules with slight differences in bonding (which might be caused by, for instance,
    # different solvent environments)
    # This tolerance was chosen based on trying to distinguish CO optimized in
    # two different solvents
    mm = MoleculeMatcher(tolerance=0.000001)

    # First, group by formula
    # Hopefully this step is unnecessary - builders should already be doing this
    for mol_key, pregroup in groupby(sorted(molecules, key=_mol_form), key=_mol_form):
        groups: list[dict[str, Any]] = list()
        for mol in pregroup:
            mol_copy = copy.deepcopy(mol)

            # Single atoms could always have identical structure
            # So grouping by geometry isn't enough
            # Need to also group by charge
            if len(mol_copy) > 1:
                mol_copy.set_charge_and_spin(0)
            matched = False

            # Group by structure
            for group in groups:
                if (
                    (mm.fit(mol_copy, group["mol"]) or mol_copy == group["mol"])
                    and mol_copy.charge == group["mol"].charge
                    and mol_copy.spin_multiplicity == group["mol"].spin_multiplicity
                ):
                    group["mol_list"].append(mol)
                    matched = True
                    break

            if not matched:
                groups.append({"mol": mol_copy, "mol_list": [mol]})

        for group in groups:
            yield group["mol_list"]


def confirm_molecule(mol: Molecule | dict):
    """
    Check that something that we expect to be a molecule is actually a Molecule
    object, and not a dictionary representation.

    :param mol (Molecule):
    :return:
    """

    if isinstance(mol, dict):
        return Molecule.from_dict(mol)
    else:
        return mol


def make_mol_graph(
    mol: Molecule, critic_bonds: list[list[int]] | None = None
) -> MoleculeGraph:
    """
    Construct a MoleculeGraph using OpenBabelNN with metal_edge_extender and
    (optionally) Critic2 bonding information.

    This bonding scheme was used to define bonding for the Lithium-Ion Battery
    Electrolyte (LIBE) dataset (DOI: 10.1038/s41597-021-00986-9)

    :param mol: Molecule to be converted to MoleculeGraph
    :param critic_bonds: (optional) List of lists [a, b], where a and b are
        atom indices (0-indexed)

    :return: mol_graph, a MoleculeGraph
    """
    mol_graph = MoleculeGraph.from_local_env_strategy(mol, OpenBabelNN())
    mol_graph = metal_edge_extender(mol_graph)
    if critic_bonds:
        mg_edges = mol_graph.graph.edges()
        for bond in critic_bonds:
            bond.sort()
            if bond[0] != bond[1]:
                bond_tup = (bond[0], bond[1])
                if bond_tup not in mg_edges:
                    mol_graph.add_edge(bond_tup[0], bond_tup[1])
    return mol_graph


def get_graph_hash(mol: Molecule, node_attr: str | None = None):
    """
    Return the Weisfeiler Lehman (WL) graph hash of the MoleculeGraph described
    by this molecule, using the OpenBabelNN strategy with extension for
    metal coordinate bonds

    :param mol: Molecule
    :param node_attr: Node attribute to be used to compute the WL hash
    :return: string of the WL graph hash
    """

    mg = make_mol_graph(mol)
    return weisfeiler_lehman_graph_hash(
        mg.graph.to_undirected(),
        node_attr=node_attr,
    )


def get_molecule_id(mol: Molecule, node_attr: str | None = None):
    """
    Return an MPculeID for a molecule, with the hash component
    based on a particular attribute of the molecule graph representation.

    :param mol: Molecule
    :param node_attr:Node attribute to be used to compute the WL hash

    :return: MPculeID
    """

    graph_hash = get_graph_hash(mol, node_attr=node_attr)
    return MPculeID(
        "{}-{}-{}-{}".format(
            graph_hash,
            mol.composition.alphabetical_formula.replace(" ", ""),
            str(int(mol.charge)).replace("-", "m"),
            str(mol.spin_multiplicity),
        )
    )


def jsanitize(obj, strict=False, allow_bson=False):
    """
    This method cleans an input json-like object, either a list or a dict or
    some sequence, nested or otherwise, by converting all non-string
    dictionary keys (such as int and float) to strings, and also recursively
    encodes all objects using Monty's as_dict() protocol.
    Args:
        obj: input json-like object.
        strict (bool): This parameters sets the behavior when jsanitize
            encounters an object it does not understand. If strict is True,
            jsanitize will try to get the as_dict() attribute of the object. If
            no such attribute is found, an attribute error will be thrown. If
            strict is False, jsanitize will simply call str(object) to convert
            the object to a string representation.
        allow_bson (bool): This parameters sets the behavior when jsanitize
            encounters an bson supported type such as objectid and datetime. If
            True, such bson types will be ignored, allowing for proper
            insertion into MongoDb databases.
    Returns:
        Sanitized dict that can be json serialized.
    """
    if allow_bson and (
        isinstance(obj, (datetime.datetime, bytes))
        or (bson is not None and isinstance(obj, bson.objectid.ObjectId))
    ):
        return obj

    if isinstance(obj, (list, tuple, set)):
        return [jsanitize(i, strict=strict, allow_bson=allow_bson) for i in obj]
    if np is not None and isinstance(obj, np.ndarray):
        return [
            jsanitize(i, strict=strict, allow_bson=allow_bson) for i in obj.tolist()
        ]
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, dict):
        return {
            k.__str__(): jsanitize(v, strict=strict, allow_bson=allow_bson)
            for k, v in obj.items()
        }
    if isinstance(obj, MSONable):
        return {
            k.__str__(): jsanitize(v, strict=strict, allow_bson=allow_bson)
            for k, v in obj.as_dict().items()
        }

    if isinstance(obj, BaseModel):
        return {
            k.__str__(): jsanitize(v, strict=strict, allow_bson=allow_bson)
            for k, v in obj.model_dump().items()
        }
    if isinstance(obj, (int, float)):
        if np.isnan(obj):
            return 0
        return obj

    if obj is None:
        return None

    if not strict:
        return obj.__str__()

    if isinstance(obj, str):
        return obj.__str__()

    return jsanitize(obj.as_dict(), strict=strict, allow_bson=allow_bson)


def utcnow() -> datetime.datetime:
    """Get UTC time right now."""
    return datetime.datetime.now(datetime.timezone.utc)


def convert_datetime(
    v: datetime.datetime | dict[str, str] | str | None,
) -> datetime.datetime:
    """Validate datetime-like objects.

    Parameters
    -----------
    v : datetime, dict[str,str], str or None

    Returns
    -----------
    datetime
    """
    if not v:
        return utcnow()

    if isinstance(v, dict):
        if v.get("$date"):
            dt = datetime.datetime.fromisoformat(v["$date"])
            if not dt.tzinfo:
                dt = dt.replace(tzinfo=datetime.timezone.utc)
            return dt

    if isinstance(v, str):
        dt = datetime.datetime.fromisoformat(v)
        if not dt.tzinfo:
            dt = dt.replace(tzinfo=datetime.timezone.utc)
        return dt

    v = MontyDecoder().process_decoded(v)
    if isinstance(v, datetime.datetime) and not v.tzinfo:
        v = v.replace(tzinfo=datetime.timezone.utc)
    return v  # type: ignore[return-value]


def get_hash_blocked(
    file_path: FSPathType, chunk_size: int = 4 * 1024 * 1024, hasher: Any | None = None
) -> str:
    """
    Get the hash of a file in byte chunks.

    Parameters
    -----------
    file_path : FSPathType
    chunk_size : int = 1,000,000 bytes (default)
        The byte chunk size to use in iteratively computing the hash.
    hahser : function to compute hashes. Defaults to blake3 if available,
        and MD5 if not.

    Returns
    -----------
    The hash as a str
    """
    if hasher is None:
        if blake3:
            hasher = blake3.blake3()
        else:
            hasher = hashlib.md5()

    with open(str(file_path), "rb") as f:
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            hasher.update(data)
        return hasher.hexdigest()


def dynamic_import(module_path: str) -> Any:
    """Import arbitrary module or object."""
    paths = module_path.split(".")
    for i in range(len(paths), 0, -1):
        try:
            ob = import_module(".".join(paths[:i]))
            for path in paths[i:]:
                ob = getattr(ob, path)
            return ob
        except Exception:
            continue
    raise ValueError(f"Could not import string:\n{module_path}")


def get_flat_models_from_model(
    model: BaseModel, known_models: set[BaseModel] = set()
) -> set[BaseModel]:
    """Get all sub-models from a pydantic model.

    Args:
        model (BaseModel): Pydantic model
        known_models (set[BaseModel]) : set of identified pydantic sub-models

    Returns:
        (set[BaseModel]): Set of pydantic models
    """
    known_models = set()

    def get_sub_models(model: Any):
        if lenient_issubclass(model, BaseModel):
            known_models.add(model)
            for field_info in model.model_fields.values():
                get_sub_models(field_info.annotation)
        else:
            for type_anno in get_args(model):
                get_sub_models(type_anno)

    get_sub_models(model)
    return known_models


def requires_arrow(func: Callable) -> Callable:
    """Decorator for pyarrow-dependent functionality."""

    def wrap(*args, **kwargs):
        if not ARROW_COMPATIBLE:
            raise ImportError(
                "You must `pip install pyarrow` to use this functionality."
            )
        return func(*args, **kwargs)

    return wrap



"""Define schemas for ion migration."""

from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING, Literal

import numpy as np
from pydantic import BaseModel, Field
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import Element, Structure
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry

from emmet.core.base import EmmetBaseModel
from emmet.core.neb import NebPathwayResult
from emmet.core.types.enums import ValueEnum
from emmet.core.utils import arrow_incompatible
from emmet.core.types.typing import DateTimeType

try:
    from pymatgen.analysis.diffusion.neb.full_path_mapper import MigrationGraph
    from pymatgen.analysis.diffusion.utils.edge_data_from_sc import (
        add_edge_data_from_sc,
    )
    from pymatgen.analysis.diffusion.utils.supercells import (
        get_sc_fromstruct,
        get_start_end_structures,
    )
except ImportError:
    raise ImportError(
        "pip install pymatgen-analysis-diffusion to use MigrationGraphDoc"
    )

if TYPE_CHECKING:
    from typing import Any

    from typing_extensions import Self


class HopState(ValueEnum):
    SUCCESS = "successful"
    FAILED = "failed"
    ERROR = "error"
    UNMATCHED = "unmatched"


class HopSummary(BaseModel):
    """Store high-level migration data."""

    hop_label: int | None = Field(None, description="The label of this hop.")

    hop_key: str | None = Field(
        None, description="The pair of indices representing this hop."
    )

    hop_distance: float | None = Field(
        None, description="The distance traversed by the migrating ion."
    )

    cost: float | None = Field(None, description="The penalty in this migration.")

    match_state: HopState | None = Field(
        None, description="Whether the hop could be matched to a migration graph hop."
    )


@arrow_incompatible
class MigrationGraphDoc(EmmetBaseModel):
    """
    MigrationGraph Doc.
    Stores MigrationGraph and info such as ComputedStructureEntries (ComputedEntry can be used for working ion)
    and cutoff distance that are used to generated the object.

    Note: this doc is not self-contained within pymatgen, as it has dependence on pymatgen.analysis.diffusion,
    a namespace package aka pymatgen-diffusion.
    """

    battery_id: str | None = Field(
        None, description="The battery id for this MigrationGraphDoc"
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation for this MigrationGraph document.",
    )

    warnings: list[str] = Field(
        [], description="Any warnings related to this property."
    )

    deprecated: bool = Field(
        False,
        description="Indicates whether a migration graph fails to be constructed from the provided entries. Defaults to False, indicating mg can be constructed from entries.",  # noqa: E501
    )

    hop_cutoff: float | None = Field(
        None,
        description="The numerical value in angstroms used to cap the maximum length of a hop.",
    )

    entries_for_generation: list[ComputedStructureEntry] | None = Field(
        None,
        description="A list of ComputedStructureEntries used to generate the structure with all working ion sites.",
    )

    working_ion_entry: ComputedEntry | ComputedStructureEntry | None = Field(
        None, description="The ComputedStructureEntry of the working ion."
    )

    migration_graph: MigrationGraph | None = Field(
        None,
        description="The MigrationGraph object as defined in pymatgen.analysis.diffusion.",
    )

    populate_sc_fields: bool = Field(
        True,
        description="Flag indicating whether this document has populated the supercell fields",
    )

    sc_gen_schema: str | None = Field(
        None,
        description="The schema used to generate supercell fields. "
        "hops_only: contains only endpoints for sc hops; complete: contains all transformed uc sites.",
    )

    min_length_sc: float | None = Field(
        None,
        description="The minimum length used to generate supercell using pymatgen.",
    )

    minmax_num_atoms: tuple[int, int] | None = Field(
        None,
        description="The min/max number of atoms used to genreate supercell using pymatgen.",
    )

    matrix_supercell_structure: Structure | None = Field(
        None,
        description=(
            "The matrix suprcell structure that does not contain the "
            "mobile ions for the purpose of migration analysis."
        ),
    )

    conversion_matrix: list[list[int | float]] | None = Field(
        None,
        description="The conversion matrix used to convert unit cell to supercell.",
    )

    inserted_ion_coords: list[dict[str, list[float] | str | int]] | None = Field(
        None,
        description="A dictionary containing all mobile ion fractional coordinates in terms of supercell.",
    )

    insert_coords_combo: list[str] | None = Field(
        None,
        description=(
            "A list of combinations 'a+b' to designate hops in the supercell. "
            "Each combo should correspond to one unique hop in MigrationGraph."
        ),
    )

    paths_summary: dict[int, list[HopSummary]] | None = Field(
        None,
        description=(
            "A dictionary of ranked intercalation pathways given cost "
            "(e.g. migration barrier from transition state calcs) of each unique_hop."
        ),
    )

    migration_graph_w_cost: MigrationGraph | None = Field(
        None,
        description="MigrationGraph instance that contains cost (e.g. from NEB) information",
    )

    @classmethod
    def from_entries_and_distance(
        cls,
        battery_id: str,
        grouped_entries: list[ComputedStructureEntry],
        working_ion_entry: ComputedEntry | ComputedStructureEntry,
        hop_cutoff: float,
        populate_sc_fields: bool = True,
        sc_gen_schema: Literal["complete", "hops_only"] = "complete",
        ltol: float = 0.2,
        stol: float = 0.3,
        angle_tol: float = 5,
        min_length_sc: float | None = None,
        minmax_num_atoms: tuple[int, int] | None = None,
        **kwargs,
    ) -> Self:
        """
        Take a list of ComputedStructureEntry and generate a MigrationGraphDoc.

        Note: Can use a ComputedEntry for the working ion.

        Then a MigrationGraph object is generated with with_distance() method with a designated cutoff.
        If populate_sc_fields set to True, this method will populate the supercell related fields.
        Required kwargs are min_length_sc and minmax_num_atoms.
        """

        ranked_structures = MigrationGraph.get_structure_from_entries(
            entries=grouped_entries, migrating_ion_entry=working_ion_entry
        )
        max_sites_struct = ranked_structures[0]

        migration_graph = MigrationGraph.with_distance(
            structure=max_sites_struct,
            migrating_specie=working_ion_entry.composition.chemical_system,
            max_distance=hop_cutoff,
        )

        if populate_sc_fields:
            if min_length_sc is None or minmax_num_atoms is None:
                raise TypeError(
                    "Ensure that min_length_sc and minmax_num_atoms are set when using populate_sc_fields = True."
                )

            sm = StructureMatcher(ltol, stol, angle_tol)
            (
                host_sc,
                sc_mat,
                min_length_sc,
                minmax_num_atoms,
                coords_list,
                combo,
            ) = cls.generate_sc_fields(
                mg=migration_graph,
                min_length_sc=min_length_sc,
                minmax_num_atoms=minmax_num_atoms,
                sm=sm,
                sc_gen_schema=sc_gen_schema,
            )

            return cls(
                battery_id=battery_id,
                hop_cutoff=hop_cutoff,
                entries_for_generation=grouped_entries,
                working_ion_entry=working_ion_entry,
                migration_graph=migration_graph,
                matrix_supercell_structure=host_sc,
                conversion_matrix=sc_mat,
                inserted_ion_coords=coords_list,
                insert_coords_combo=combo,
                sc_gen_schema=sc_gen_schema,
                **kwargs,
            )

        return cls(
            battery_id=battery_id,
            hop_cutoff=hop_cutoff,
            entries_for_generation=grouped_entries,
            working_ion_entry=working_ion_entry,
            migration_graph=migration_graph,
            **kwargs,
        )

    @classmethod
    def augment_from_mgd_and_npr(
        cls,
        mgd: "MigrationGraphDoc",
        npr: NebPathwayResult,
        barrier_type: Literal["max_barrier", "energy_range"],
    ) -> Self:
        """
        Takes an existing MigrationGraphDoc and augment it.

        Specifically, the `paths_summary` and `migration_graph_w_cost` fields
        will be populated with transition state data from NebPathwayResult.

        `barrier_type` can be set to 'barrier' or 'energy_range'.
        See docstring of get_paths_summary_with_neb_res for detail.
        """
        mgd_w_cost = copy.deepcopy(mgd)
        if not mgd_w_cost.migration_graph:
            raise ValueError("A MigrationGraph must be provided to augment it.")
        paths_summary, mg_new = cls.get_paths_summary_with_neb_res(
            mg=mgd_w_cost.migration_graph, npr=npr, barrier_type=barrier_type
        )
        mgd_w_cost.paths_summary = paths_summary
        mgd_w_cost.migration_graph_w_cost = mg_new
        return mgd_w_cost

    @staticmethod
    def generate_sc_fields(
        mg: MigrationGraph,
        min_length_sc: float,
        minmax_num_atoms: tuple[int, int],
        sm: StructureMatcher,
        sc_gen_schema: Literal["hops_only", "complete"],
    ):
        if sc_gen_schema not in ["hops_only", "complete"]:
            raise ValueError(
                f"Invalid sc_gen_schema: {sc_gen_schema}. Specify 'hops_only' or 'complete'."
            )

        min_length_sc = min_length_sc
        minmax_num_atoms = minmax_num_atoms

        sc_mat = get_sc_fromstruct(
            base_struct=mg.structure,
            min_atoms=minmax_num_atoms[0],
            max_atoms=minmax_num_atoms[1],
            min_length=min_length_sc,
        )

        sc_mat = sc_mat.tolist()  # type: ignore[attr-defined]
        host_sc = mg.host_structure * sc_mat
        working_ion = mg.only_sites[0].species_string

        if sc_gen_schema == "hops_only":
            coords_list, combo = MigrationGraphDoc.get_sc_info_hops_only(mg, sc_mat)

        if sc_gen_schema == "complete":
            coords_list = MigrationGraphDoc.ordered_sc_site_list(mg.only_sites, sc_mat)
            combo, coords_list = MigrationGraphDoc.get_hop_sc_combo(
                mg.unique_hops, sc_mat, sm, host_sc, working_ion, coords_list
            )

        return host_sc, sc_mat, min_length_sc, minmax_num_atoms, coords_list, combo

    @staticmethod
    def get_sc_info_hops_only(mg: MigrationGraph, sc_mat: list[list[int | float]]):
        coords_list = []
        combo = []
        base_struct = mg.host_structure

        for one_hop in mg.unique_hops.values():
            migration_hop = one_hop["hop"]
            isite, iindex = migration_hop.isite, one_hop["iindex"]
            esite, eindex = migration_hop.esite, one_hop["eindex"]
            start_structure, end_structure, _ = get_start_end_structures(
                isite=isite,
                esite=esite,
                base_struct=base_struct,
                sc_mat=sc_mat,
                vac_mode=False,
            )

            sc_ini_info = {
                "uc_site_type": iindex,
                "site_frac_coords": tuple(start_structure[0].frac_coords),
            }
            if sc_ini_info not in coords_list:
                coords_list.append(sc_ini_info)
            combo_ini = coords_list.index(sc_ini_info)

            sc_fin_info = {
                "uc_site_type": eindex,
                "site_frac_coords": tuple(end_structure[0].frac_coords),
            }
            if sc_fin_info not in coords_list:
                coords_list.append(sc_fin_info)
            combo_fin = coords_list.index(sc_fin_info)

            combo.append(f"{combo_ini}+{combo_fin}")

        for coords in coords_list:
            coords["site_frac_coords"] = list(coords["site_frac_coords"])

        return coords_list, combo

    @staticmethod
    def ordered_sc_site_list(uc_sites_only: Structure, sc_mat: list[list[int]]):
        uc_no_site = uc_sites_only.copy()
        uc_no_site.remove_sites(range(len(uc_sites_only)))
        working_ion = uc_sites_only[0].species_string
        sc_site_dict = {}  # type: dict

        for i, e in enumerate(uc_sites_only):
            uc_one_set = uc_no_site.copy()
            uc_one_set.insert(0, working_ion, e.frac_coords)
            sc_one_set = uc_one_set * sc_mat
            for index in range(len(sc_one_set)):
                sc_site_dict[len(sc_site_dict) + 1] = {
                    "uc_site_type": i,
                    "site_frac_coords": list(sc_one_set[index].frac_coords),
                    # "extra_site": False
                }

        ordered_site_list = [
            e
            for i, e in enumerate(
                sorted(
                    sc_site_dict.values(),
                    key=lambda v: float(np.linalg.norm(v["site_frac_coords"])),
                )
            )
        ]
        return ordered_site_list

    @staticmethod
    def get_hop_sc_combo(
        unique_hops: dict,
        sc_mat: list[list[int]],
        sm: StructureMatcher,
        host_sc: Structure,
        working_ion: str,
        ordered_sc_site_list: list,
    ) -> tuple[list[str], list[dict[str, Any]]]:
        combo = []

        unique_hops = {k: v for k, v in sorted(unique_hops.items())}
        for one_hop in unique_hops.values():
            added = False
            sc_isite_set = {
                k: v
                for k, v in enumerate(ordered_sc_site_list)
                if v["uc_site_type"] == one_hop["iindex"]
            }
            sc_esite_set = {
                k: v
                for k, v in enumerate(ordered_sc_site_list)
                if v["uc_site_type"] == one_hop["eindex"]
            }
            for sc_iindex, sc_isite in sc_isite_set.items():
                for sc_eindex, sc_esite in sc_esite_set.items():
                    sc_check = host_sc.copy()
                    sc_check.insert(0, working_ion, sc_isite["site_frac_coords"])
                    sc_check.insert(1, working_ion, sc_esite["site_frac_coords"])
                    if MigrationGraphDoc.compare_sc_one_hop(
                        one_hop,
                        sc_mat,
                        sm,
                        host_sc,
                        sc_check,
                        working_ion,
                        (sc_isite["uc_site_type"], sc_esite["uc_site_type"]),
                    ):
                        combo.append(f"{sc_iindex}+{sc_eindex}")
                        added = True
                        break
                if added:
                    break

            if not added:
                new_combo, ordered_sc_site_list = MigrationGraphDoc.append_new_site(
                    host_sc, ordered_sc_site_list, one_hop, sc_mat, working_ion
                )
                combo.append(new_combo)

        return combo, ordered_sc_site_list

    @staticmethod
    def compare_sc_one_hop(
        one_hop: dict,
        sc_mat: list,
        sm: StructureMatcher,
        host_sc: Structure,
        sc_check: Structure,
        working_ion: str,
        uc_site_types: tuple[int, int],
    ) -> bool:
        sc_mat_inv = np.linalg.inv(sc_mat)
        convert_sc_icoords = np.dot(one_hop["ipos"], sc_mat_inv)
        convert_sc_ecoords = np.dot(one_hop["epos"], sc_mat_inv)
        convert_sc = host_sc.copy()
        convert_sc.insert(0, working_ion, convert_sc_icoords)
        convert_sc.insert(1, working_ion, convert_sc_ecoords)

        if sm.fit(convert_sc, sc_check):
            one_hop_dis = one_hop["hop"].length
            sc_check_hop_dis = np.linalg.norm(sc_check[0].coords - sc_check[1].coords)
            if np.isclose(one_hop_dis, sc_check_hop_dis, atol=1e-5):
                if (
                    one_hop["iindex"] == uc_site_types[0]
                    and one_hop["eindex"] == uc_site_types[1]
                ):
                    return True

        return False

    @staticmethod
    def append_new_site(
        host_sc: Structure,
        ordered_sc_site_list: list,
        one_hop: dict,
        sc_mat: list[list[int]],
        working_ion: str,
    ) -> tuple[str, list[dict[str, Any]]]:
        sc_mat_inv = np.linalg.inv(sc_mat)
        sc_ipos = np.dot(one_hop["ipos"], sc_mat_inv)
        sc_epos = np.dot(one_hop["epos"], sc_mat_inv)
        sc_iindex, sc_eindex = None, None
        host_sc_insert = host_sc.copy()

        for k, v in enumerate(ordered_sc_site_list):
            if np.allclose(sc_ipos, v["site_frac_coords"], rtol=0.1, atol=0.1):
                sc_iindex = k
            if np.allclose(sc_epos, v["site_frac_coords"], rtol=0.1, atol=0.1):
                sc_eindex = k

        if sc_iindex is None:
            host_sc_insert.insert(0, working_ion, sc_ipos)
            ordered_sc_site_list.append(
                {
                    "uc_site_type": one_hop["iindex"],
                    "site_frac_coords": list(host_sc_insert[0].frac_coords),
                    "extra_site": True,
                }
            )
            sc_iindex = len(ordered_sc_site_list) - 1
        if sc_eindex is None:
            host_sc_insert.insert(0, working_ion, sc_epos)
            ordered_sc_site_list.append(
                {
                    "uc_site_type": one_hop["eindex"],
                    "site_frac_coords": list(host_sc_insert[0].frac_coords),
                    "extra_site": True,
                }
            )
            sc_eindex = len(ordered_sc_site_list) - 1

        return f"{sc_iindex}+{sc_eindex}", ordered_sc_site_list

    @staticmethod
    def get_distinct_hop_sites(
        inserted_ion_coords: list[dict[str, list[float]]],
        insert_coords_combo: list[str],
    ) -> tuple[list[list[float]], list[str], dict[str, str]]:
        """
        Expand the list of inserted coordinates and hop indices.

        Parameters
        -----------
        inserted_ion_coords: list[dict[str, list[float]]]
            List of dict's containing the working ion coordinates.
        insert_coords_combo: list[str]
            List of hop indices of the form "<int>+<int>"

        Returns
        -----------
        list of list of float
            List of working ion coordinates
        list of str
            List of hop keys
        dict of str to str
            Dict mapping of hop keys in different formats
        """
        dis_sites_list: list[list[float]] = []
        dis_combo_list: list[str] = []
        mgdoc_sites_mapping: dict[int, int] = {}
        combo_mapping = {}

        for one_combo in insert_coords_combo:
            ini, end = list(map(int, one_combo.split("+")))

            if ini in mgdoc_sites_mapping.keys():
                dis_ini = mgdoc_sites_mapping[ini]
            else:
                dis_sites_list.append(
                    list(inserted_ion_coords[ini]["site_frac_coords"])
                )
                dis_ini = len(dis_sites_list) - 1
                mgdoc_sites_mapping[ini] = dis_ini
            if end in mgdoc_sites_mapping.keys():
                dis_end = mgdoc_sites_mapping[end]
            else:
                dis_sites_list.append(
                    list(inserted_ion_coords[end]["site_frac_coords"])
                )
                dis_end = len(dis_sites_list) - 1
                mgdoc_sites_mapping[end] = dis_end

            dis_combo = f"{dis_ini}+{dis_end}"
            dis_combo_list.append(dis_combo)
            combo_mapping[dis_combo] = one_combo

        return dis_sites_list, dis_combo_list, combo_mapping

    @staticmethod
    def get_paths_summary_with_neb_res(
        mg: MigrationGraph,
        npr: NebPathwayResult,
        barrier_type: Literal["max_barrier", "energy_range"],
        zero_short_hop_cost: bool = True,
        verbose: bool = False,
    ) -> tuple[dict[int, list[HopSummary]], MigrationGraph]:
        """
        This is a post-processing function that matches the results of transition state cals (NEB or ApproxNEB)
        and unique_hops in the MigrationGraph, and then outputs a ranked list according to the calculated barrier

        Parameters
        ----------
        mgd: MigrationGraph
            The MigrationGraph to be matched and get paths from
        npr: NebPathwayResult
            The doc used to get transition state calc info
        barrier_type: str
            The type of barrier used to assign cost. Currently supporting
            'barrier': max of forward & reverse barrier (max energy - ep energy)
            'energy_range': max of energies - min of energies
        """
        if barrier_type not in ["max_barrier", "energy_range"]:
            raise ValueError(
                f"Invalid barrier_type: {barrier_type}. Specify 'max_barrier' or 'energy_range'."
            )

        energy_struct_info = MigrationGraphDoc._get_energy_struct_info(npr)
        mg_new = copy.deepcopy(mg)

        for info in energy_struct_info.values():
            try:
                add_edge_data_from_sc(
                    mg_new,
                    i_sc=info["input_endpoints"][0],
                    e_sc=info["input_endpoints"][-1],
                    data_array=info,  # type: ignore[arg-type]
                    key="energy_struct_info",
                )
            except (RuntimeError, ValueError) as e:
                if verbose:
                    logging.warning(f"{e} occured to during matching")

        unmatched_uhops = []
        failed_neb_uhops = []

        for k, v in mg_new.unique_hops.items():
            # if a unique_hop failed to match with an NEB hops calc (no symmetry match, missing calc, etc.)
            # the cost is set to infinity.
            # if a unique_hops matches to an NEB hop in the FAILED state
            # cost is set to the insertion energy diff (barrier from NebResult)
            # but the failed state is explicitly stated in paths_summary
            if "energy_struct_info" not in v:
                cost = float("inf")
                hop_key = ""
                state = "unmatched"
                unmatched_uhops.append(k)
            else:
                state = v["energy_struct_info"]["state"]
                if state == HopState.FAILED:
                    failed_neb_uhops.append(k)
                cost, hop_key = (
                    v["energy_struct_info"][barrier_type],
                    v["energy_struct_info"]["hop_key"],
                )
                # for short hops with low barrier, set cost to 0
                if zero_short_hop_cost:
                    cutoff = 2 * MigrationGraphDoc._get_wi_ionic_radius(mg_new)
                    if MigrationGraphDoc._check_short_hop(
                        v, current_cost=cost, length_cutoff=cutoff
                    ):
                        cost = 0
            mg_new.add_data_to_similar_edges(
                target_label=v["hop_label"],
                data={"cost": cost, "hop_key": hop_key, "match_state": state},
            )

        if verbose:
            if unmatched_uhops:
                logging.warning(
                    f"The following unique hops have not matched: {unmatched_uhops}"
                )
            if failed_neb_uhops:
                logging.warning(
                    f"The following unique_hops matched but the corresponding calculation "
                    f"is in the FAILED state: {failed_neb_uhops}"
                )

        paths = list(mg_new.get_path())
        paths_summary = {}
        for one_path in paths:
            path_summary = []
            for one_hop in one_path[1]:
                hop_summary = {
                    k: one_hop.get(k, None)
                    for k in (
                        "hop_label",
                        "hop_key",
                        "hop_distance",
                        "cost",
                        "match_state",
                    )
                }
                if isinstance(hs := hop_summary.get("match_state"), str):
                    hop_summary["match_state"] = HopState(hs)
                path_summary.append(HopSummary(**hop_summary))

            paths_summary[one_path[0]] = path_summary

        return paths_summary, mg_new

    @staticmethod
    def _get_energy_struct_info(npr: NebPathwayResult) -> dict[str, dict[str, Any]]:
        """
        Convert results in an NebPathWayResult object to a dict.

        This method exists primarily to populate MigrationGraph.unique_hops.
        """
        energy_struct_info = {}
        for hop_key, data in npr.hops.items():
            energy_struct_info[hop_key] = {
                "hop_key": hop_key,
                "max_barrier": npr.max_barriers[hop_key],
                "energy_range": data.barrier_energy_range,
                "energies": data.energies,
                "state": data.state,
                "calc_fail_info": data.failure_reasons,
                "input_endpoints": [data.initial_images[0], data.initial_images[-1]],
                "output_structs": data.images,
            }
        return energy_struct_info

    @staticmethod
    def _get_wi_ionic_radius(mg: MigrationGraph) -> float:
        """
        Guess the ionic radius of the working ion from a MigrationGraph.
        """
        wi = Element(mg.only_sites[0].species.elements[0].name)  # type: ignore[arg-type]
        if (
            wi_oxi_state_guess := mg.structure.composition.oxi_state_guesses()[0][
                wi.name
            ]
        ) in wi.ionic_radii:
            return wi.ionic_radii[wi_oxi_state_guess]  # type: ignore[index]
        return wi.ionic_radii[min(wi.ionic_radii)]

    @staticmethod
    def _check_short_hop(
        unique_hop: dict[str, Any],
        current_cost: float,
        length_cutoff: float,
        barrier_cutoff: float = 0.02,
    ) -> bool:
        """
        Check whether a migration hop is within distance and barrier cutoffs.

        Parameters
        -----------
        unique_hop : dict of str to Any
            Representation of the hop, including distance
        current_cost : float
            Penalty associated with the migration event
        length_cutoff : float
            Distance cutoff for a single hop
        barrier_cutoff : float, default 0.02
            Cutoff for the energy barrier of the hop.
        """
        hop_length = unique_hop["hop_distance"]
        if hop_length < length_cutoff and current_cost < barrier_cutoff:
            return True
        return False



from hashlib import blake2b

from pydantic import Field

from emmet.core.molecules import MolPropertyOrigin
from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPculeID
from emmet.core.qchem.task import TaskDocument

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"

CHARGES_METHODS = ["nbo", "resp", "critic2", "mulliken"]
SPINS_METHODS = ["nbo", "mulliken"]


class PartialChargesDoc(PropertyDoc):
    """Atomic partial charges of a molecule"""

    property_name: str = "partial_charges"

    method: str = Field(
        ..., description="Method used to compute atomic partial charges"
    )

    partial_charges: list[float] = Field(
        ..., description="Atomic partial charges for the molecule"
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        preferred_methods: list[str],
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Determine partial charges from a task document

        :param task: task document from which partial charges can be extracted
        :param molecule_id: MPculeID
        :param preferred_methods: list of methods; by default, NBO7, RESP, Critic2, and Mulliken, in that order
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        charges = None
        method = None

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        for m in preferred_methods:
            if m == "nbo" and task.output.nbo is not None:
                method = m
                charges = [
                    float(task.output.nbo["natural_populations"][0]["Charge"][str(i)])
                    for i in range(len(mol))
                ]
                break
            elif m == "resp" and task.output.resp is not None:
                method = m
                charges = [float(i) for i in task.output.resp]
                break
            elif m == "critic2" and task.critic2 is not None:
                method = m
                charges = list([float(i) for i in task.critic2["processed"]["charges"]])
                break
            elif m == "mulliken" and task.output.mulliken is not None:
                method = m
                if isinstance(task.output.mulliken[0], list):
                    charges = [float(mull[0]) for mull in task.output.mulliken]
                else:
                    charges = [float(i) for i in task.output.mulliken]
                break

        id_string = (
            f"partial_charges-{molecule_id}-{task.task_id}-{task.lot_solvent}-{method}"
        )
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        if charges is None:
            raise Exception("No valid partial charge information!")

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            partial_charges=charges,
            method=method,
            origins=[MolPropertyOrigin(name="partial_charges", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )


class PartialSpinsDoc(PropertyDoc):
    """Atomic partial charges of a molecule"""

    property_name: str = "partial_spins"

    method: str = Field(..., description="Method used to compute atomic partial spins")

    partial_spins: list[float] = Field(
        ..., description="Atomic partial spins for the molecule"
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        preferred_methods: list[str],
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Determine partial spins from a task document

        :param task: task document from which partial spins can be extracted
        :param molecule_id: MPculeID
        :param preferred_methods: list of methods; by default, NBO7 and Mulliken, in that order
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        spins = None
        method = None

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        if mol.spin_multiplicity == 1:
            raise Exception("Closed-shell molecule has no partial spins!")

        for m in preferred_methods:
            if m == "nbo" and task.output.nbo is not None:
                method = m
                spins = [
                    float(task.output.nbo["natural_populations"][0]["Density"][str(i)])
                    for i in range(len(mol))
                ]
                break
            elif m == "mulliken" and task.output.mulliken is not None:
                method = m
                spins = [float(mull[1]) for mull in task.output.mulliken]
                break

        id_string = (
            f"partial_spins-{molecule_id}-{task.task_id}-{task.lot_solvent}-{method}"
        )
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        if spins is None:
            raise Exception("No valid partial spin information!")

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            partial_spins=spins,
            method=method,
            origins=[MolPropertyOrigin(name="partial_spins", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )



from typing import Annotated, TypeVar

from pydantic import BeforeValidator, WrapSerializer
from pymatgen.core import Element
from typing_extensions import TypedDict

from emmet.core.types.pymatgen_types.structure_adapter import (
    TypedStructureDict,
    pop_empty_structure_keys,
)

try:
    from pymatgen.analysis.alloys.core import AlloyMember, AlloyPair, AlloySystem
except ImportError:
    raise ImportError(
        "Install pymatgen-analysis-alloys to use AlloyPairDoc or AlloySystemDoc"
    )


class TypedSupportedPropertiesDict(TypedDict):
    energy_above_hull: float
    formation_energy_per_atom: float
    band_gap: float
    is_gap_direct: bool
    m_n: float
    m_p: float
    theoretical: bool
    is_metal: bool


class TypedAlloyMemberDict(TypedDict):
    id_: str
    db: str
    composition: dict[Element, float]
    x: float
    is_ordered: bool


TypedAlloyPairDict = TypedDict(
    "TypedAlloyPairDict",
    {
        "@module": str,
        "@class": str,
        "@version": str,
        "formula_a": str,
        "formula_b": str,
        "structure_a": TypedStructureDict,
        "structure_b": TypedStructureDict,
        "id_a": str,
        "id_b": str,
        "chemsys": str,
        "alloying_element_a": str,
        "alloying_element_b": str,
        "alloying_species_a": list[str],
        "alloying_species_b": list[str],
        "observer_elements": list[str],
        "observer_species": list[str],
        "anions_a": list[str],
        "anions_b": list[str],
        "cations_a": list[str],
        "cations_b": list[str],
        "lattice_parameters_a": list[float],
        "lattice_parameters_b": list[float],
        "volume_cube_root_a": float,
        "volume_cube_root_b": float,
        "properties_a": TypedSupportedPropertiesDict,
        "properties_b": TypedSupportedPropertiesDict,
        "spacegroup_intl_number_a": int,
        "spacegroup_intl_number_b": int,
        "pair_id": str,
        "pair_formula": str,
        "alloy_oxidation_state": int,
        "isoelectronic": bool,
        "anonymous_formula": str,
        "nelements": int,
        "members": list[TypedAlloyMemberDict],
    },
)

TypedAlloySystemDict = TypedDict(
    "TypedAlloySystemDict",
    {
        "@class": str,
        "@module": str,
        "@version": str,
        "alloy_id": str,
        "alloy_pairs": list[TypedAlloyPairDict],
        "chemsys": str,
        "chemsys_size": int,
        "has_members": bool,
        "members": list[TypedAlloyMemberDict],
        "additional_members": list[TypedAlloyMemberDict],
        "n_pairs": int,
        "ids": list[str],
        "pair_ids": list[str],
    },
)


AlloyMemberTypeVar = TypeVar("AlloyMemberTypeVar", AlloyMember, TypedAlloyMemberDict)
AlloyPairTypeVar = TypeVar("AlloyPairTypeVar", AlloyPair, TypedAlloyPairDict)
AlloySystemTypeVar = TypeVar("AlloySystemTypeVar", AlloySystem, TypedAlloySystemDict)


def pop_empty_alloy_pair_structure_keys(alloy_pair: AlloyPairTypeVar):
    if isinstance(alloy_pair, dict):
        for key in ["structure_a", "structure_b"]:
            alloy_pair[key] = pop_empty_structure_keys(alloy_pair[key])  # type: ignore[literal-required]

    return alloy_pair


def pop_empty_alloy_system_structure_keys(alloy_system: AlloySystemTypeVar):
    if isinstance(alloy_system, dict):
        alloy_system["alloy_pairs"] = [
            pop_empty_alloy_pair_structure_keys(alloy_pair)
            for alloy_pair in alloy_system["alloy_pairs"]
        ]

    return alloy_system


AlloyPairType = Annotated[
    AlloyPairTypeVar,
    BeforeValidator(pop_empty_alloy_pair_structure_keys),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedAlloyPairDict),
]
AlloySystemType = Annotated[
    AlloySystemTypeVar,
    BeforeValidator(pop_empty_alloy_system_structure_keys),
    WrapSerializer(lambda x, nxt, info: x.as_dict(), return_type=TypedAlloySystemDict),
]



from pydantic import BaseModel, Field


class ElasticityData(BaseModel):
    """
    Elastic tensors and associated information.
    """

    k_voigt: float | None = Field(
        None,
        description="Voigt average of the bulk modulus in GPa.",
    )

    k_reuss: float | None = Field(
        None,
        description="Reuss average of the bulk modulus in GPa.",
    )

    k_vrh: float | None = Field(
        None,
        description="Voigt-Reuss-Hill average of the bulk modulus in GPa.",
    )

    g_voigt: float | None = Field(
        None,
        description="Voigt average of the shear modulus in GPa.",
    )

    g_reuss: float | None = Field(
        None,
        description="Reuss average of the shear modulus in GPa.",
    )

    g_vrh: float | None = Field(
        None,
        description="Voigt-Reuss-Hill average of the shear modulus in GPa.",
    )

    universal_anisotropy: float | None = Field(
        None,
        description="Elastic anisotropy.",
    )

    homogeneous_poisson: float | None = Field(
        None,
        description="Poisson's ratio.",
    )

    elastic_tensor: list[list[float]] | None = Field(
        None,
        description="Stiffness tensor in GPa.",
    )

    compliance_tensor: list[list[float]] | None = Field(
        None,
        description="Compliance tensor in 10^(-12)/Pa.",
    )


class ElasticityDoc(BaseModel):
    """
    Model for a document containing elasticity data
    """

    pretty_formula: str | None = Field(
        None,
        description="Cleaned representation of the material formula",
    )

    chemsys: str | None = Field(
        None,
        description="Dash-delimited string of elements in the material.",
    )

    elasticity: ElasticityData | None = Field(
        None,
        description="Elasticity data for the material.",
    )

    task_id: str | None = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******.",
    )



"""Module to define various calculation types as Enums for VASP."""

from __future__ import annotations
from importlib.resources import files as import_resource_files
from itertools import product

from ruamel.yaml import YAML

_BASE_ENUM_PATH = {
    "vasp": import_resource_files("emmet.core.vasp"),
    "qchem": import_resource_files("emmet.core.qchem"),
}

for code_base in _BASE_ENUM_PATH:
    _BASE_ENUM_PATH[code_base] /= "calc_types"


def get_enum_source(
    enum_name: str, doc: str, members: dict, enum_class: str = "ValueEnum"
) -> str:
    """Write python-format enum from a dict of members and metadata.

    Parameters
    -----------
    enum_name : str
        Name of the enum.
    doc : str
        Enum docstr
    members : dict
        The key-value pair indexed members of the enum.
    enum_class : str = "ValueEnum"
        The name of the enum class.

    Returns
    --------
    str
    """
    items = [f"class {enum_name}({enum_class}):", f'    """ {doc} """\n']
    items += [f'    {const} = "{val}"' for const, val in members.items()]
    return "\n".join(items)


def string_bulk_replace(string: str, rules: dict[str, str]) -> str:
    """Perform multiple string replacements subject to a set of rules.

    Parameters
    -----------
    string : str
        input string
    rules : dict[str,str]
        A dict of string replacements, with the characters to replace
        as keys, and their replacements as values.

    Returns
    --------
    str
    """
    for targ_char, rep_char in rules.items():
        string = string.replace(targ_char, rep_char)
    return string


def generate_vasp_enums_file(enum_file_name: str | None = None) -> None:
    """
    Generate VASP enum members from reference yaml data.

    Parameters
    -----------
    enum_file_name : str
        Name of the file to write the enums to.
        Defaults to _BASE_ENUM_PATH / vasp_enums.json.gz
    """

    with open(_BASE_ENUM_PATH["vasp"] / "calc_types.yaml", "r") as config:
        _RUN_TASK_TYPE_DATA = YAML().load(config)

    _TASK_TYPES = _RUN_TASK_TYPE_DATA.get("TASK_TYPES")

    _RUN_TYPES = set(
        rt
        for functionals in _RUN_TASK_TYPE_DATA.get("RUN_TYPES", {}).values()
        for rt in functionals
    )
    _RUN_TYPES.update(set(f"{rt}+U" for rt in _RUN_TYPES))

    _ENUMS = {
        "RunType": {
            "_".join(rt.split()).replace("+", "_").replace("-", "_"): rt
            for rt in _RUN_TYPES
        },
        "TaskType": {"_".join(tt.split()): tt for tt in _TASK_TYPES},
        "CalcType": {
            f"{'_'.join(rt.split()).replace('+','_').replace('-','_')}"
            f"_{'_'.join(tt.split())}": f"{rt} {tt}"
            for rt, tt in product(_RUN_TYPES, _TASK_TYPES)
        },
    }

    docstr = {}
    for enum_name in _ENUMS:
        rtc_type = enum_name.split("Calc")[-1].split("Type")[0].lower()
        if len(rtc_type) > 0:
            rtc_type += " "
        docstr[enum_name] = f"VASP calculation {rtc_type}types."

    enum_file_name = enum_file_name or str(_BASE_ENUM_PATH["vasp"] / "enums.py")
    with open(enum_file_name, "w+") as f:
        f.write(
            """\"\"\"
Autogenerated Enums for VASP RunType, TaskType, and CalcType.

Do not edit this by hand to add or remove enums.
Instead, edit
    dev_scripts/generate_enums.py
and/or
    emmet/core/vasp/calc_types/calc_types.yaml
\"\"\"
from emmet.core.types.enums import ValueEnum, IgnoreCaseEnum

"""
        )
        enum_order = (
            "RunType",
            "TaskType",
            "CalcType",
        )
        for ienum, enum_name in enumerate(enum_order):
            sorted_enums = {k: _ENUMS[enum_name][k] for k in sorted(_ENUMS[enum_name])}
            f.write(
                get_enum_source(
                    enum_name,
                    docstr[enum_name],
                    sorted_enums,
                    enum_class=(
                        "IgnoreCase"
                        if enum_name in ["RunType", "CalcType"]
                        else "Value"
                    )
                    + "Enum",
                )
            )
            f.write("\n\n" if ienum < (len(enum_order) - 1) else "\n")


def generate_qchem_enum_file(enum_file_name: str | None = None) -> None:
    """
    Generate QChem enum members from reference yaml data.

    Original author, Evan Spotte-Smith <ewcspottesmith@lbl.gov>

    Parameters
    -----------
    enum_file_name : str
        Name of the file to write the enums to.
        Defaults to _BASE_ENUM_PATH / qchem_enums.json.gz
    """

    with open(_BASE_ENUM_PATH["qchem"] / "calc_types.yaml", "r") as config:
        _calc_type_meta = YAML().load(config)

    _calc_type_meta["FUNCTIONALS"] = [
        rt
        for functionals in _calc_type_meta["FUNCTIONAL_CLASSES"].values()
        for rt in functionals
    ]

    _LOTS = list()

    for funct in _calc_type_meta["FUNCTIONALS"]:
        for basis in _calc_type_meta["BASIS_SETS"]:
            for solv_model in _calc_type_meta["SOLVENT_MODELS"]:
                _LOTS.append(f"{funct}/{basis}/{solv_model}")

    _lot_str_replacements = {
        "+": "_",
        "-": "_",
        "(": "_",
        ")": "_",
        "/": "_",
        "*": "_d",
    }

    _ENUMS = {
        "LevelOfTheory": {
            "_".join(string_bulk_replace(lot, _lot_str_replacements).split()): lot
            for lot in _LOTS
        },
        "TaskType": {
            "_".join(tt.split()).replace("-", "_"): tt
            for tt in _calc_type_meta["TASK_TYPES"]
        },
        "CalcType": {
            (
                "_".join(string_bulk_replace(lot, _lot_str_replacements).split())
                + f"_{'_'.join(tt.split()).replace('-', '_')}"
            ): f"{lot} {tt}"
            for lot, tt in product(_LOTS, _calc_type_meta["TASK_TYPES"])
        },
    }

    docstr = {
        "LevelOfTheory": "Levels of theory for calculations in Q-Chem.",
        "TaskType": "Calculation task types for Q-Chem.",
        "CalcType": "Calculation types (LOT + task type) for Q-Chem.",
    }

    enum_file_name = enum_file_name or str(_BASE_ENUM_PATH["qchem"] / "enums.py")

    with open(enum_file_name, "w+") as f:
        f.write(
            """\"\"\"
Autogenerated Enums for Q-Chem LevelOfTheory, TaskType, and CalcType.

Do not edit this by hand to add or remove enums.
Instead, edit
    dev_scripts/generate_enums.py
and/or
    emmet/core/qchem/calc_types/calc_types.yaml
\"\"\"
from emmet.core.types.enums import ValueEnum

"""
        )
        enum_order = (
            "LevelOfTheory",
            "TaskType",
            "CalcType",
        )
        for ienum, enum_name in enumerate(enum_order):
            sorted_enums = {k: _ENUMS[enum_name][k] for k in sorted(_ENUMS[enum_name])}
            f.write(get_enum_source(enum_name, docstr[enum_name], sorted_enums))
            f.write("\n\n" if ienum < (len(enum_order) - 1) else "\n")


if __name__ == "__main__":
    generate_vasp_enums_file()
    generate_qchem_enum_file()



import json
import datetime

import pytest

from monty.io import zopen

from emmet.core.qchem.task import TaskDocument
from emmet.core.molecules.vibration import VibrationDoc


@pytest.fixture(scope="session")
def test_tasks(test_dir):
    with zopen(test_dir / "liec_tasks.json.gz") as f:
        data = json.load(f)

    for d in data:
        d["last_updated"] = datetime.datetime.strptime(
            d["last_updated"]["string"], "%Y-%m-%d %H:%M:%S.%f"
        )

    tasks = [TaskDocument(**t) for t in data]
    return tasks


@pytest.fixture(scope="session")
def raman_task(test_dir):
    with zopen(test_dir / "raman_task.json.gz") as f:
        data = json.load(f)

    data["last_updated"] = datetime.datetime.strptime(
        data["last_updated"]["string"], "%Y-%m-%d %H:%M:%S.%f"
    )

    task = TaskDocument(**data)
    return task


def test_vibration(test_tasks):
    task = test_tasks[0]

    vib_doc = VibrationDoc.from_task(
        task, molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )
    assert vib_doc.property_name == "vibrations"
    assert len(vib_doc.frequencies) == 27
    assert len(vib_doc.frequency_modes) == 27
    assert len(vib_doc.ir_intensities) == 27
    assert vib_doc.frequencies[0] == pytest.approx(49.47)
    assert vib_doc.ir_intensities[0] == pytest.approx(93.886)
    assert vib_doc.ir_activities[0] is True


def test_raman(raman_task):
    vib_doc = VibrationDoc.from_task(
        raman_task, molecule_id="b9ba54febc77d2a9177accf4605767db-C1Li2O3-1-2"
    )

    assert vib_doc.raman_activities is not None
    assert vib_doc.raman_intensities[0] == pytest.approx(0.424)



from datetime import datetime

from pydantic import BaseModel, Field
from typing_extensions import TypedDict


class TypedUserSettingsDict(TypedDict):
    institution: str
    sector: str
    job_role: str
    is_email_subscribed: bool
    message_last_read: datetime
    agreed_terms: bool


class UserSettingsDoc(BaseModel):
    """
    Defines data for user settings
    """

    consumer_id: str | None = Field(
        None, title="Consumer ID", description="Consumer ID for a specific user."
    )

    settings: TypedUserSettingsDict | None = Field(
        None,
        title="Consumer ID settings",
        description="Settings defined for a specific user.",
    )



from hashlib import blake2b

from pydantic import Field

from emmet.core.molecules import MolPropertyOrigin
from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPculeID
from emmet.core.qchem.task import TaskDocument
from emmet.core.types.pymatgen_types.structure_adapter import MoleculeType

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class VibrationDoc(PropertyDoc):
    property_name: str = "vibrations"

    molecule: MoleculeType = Field(..., description="Molecular structure")

    frequencies: list[float] = Field(
        ..., description="List of molecular vibrational frequencies"
    )

    frequency_modes: list[list[list[float]]] = Field(
        ..., description="Vibrational frequency modes of the molecule"
    )

    ir_intensities: list[float] = Field(
        ...,
        title="IR intensities",
        description="Intensities for IR vibrational spectrum peaks",
    )

    ir_activities: list[bool] = Field(
        ...,
        title="IR activities",
        description="List indicating if frequency-modes are IR-active",
    )

    raman_intensities: list[float] | None = Field(
        None,
        title="Raman intensities",
        description="Intensities for Raman spectrum peaks",
    )

    raman_activities: list[float] | None = Field(
        None,
        title="Raman activities",
        description="List indicating if frequency-modes are Raman-active",
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Construct a vibration document from a task document

        :param task: document from which vibrational properties can be extracted
        :param molecule_id: MPculeID
        :param deprecated: bool. Is this document deprecated?
        :param kwargs: to pass to PropertyDoc
        :return:
        """

        if task.output.frequencies is None:
            raise Exception("No frequencies in task!")

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        frequencies = task.output.frequencies
        frequency_modes = None
        ir_intensities = None
        ir_active = None
        raman_intensities = None
        raman_active = None
        for calc in task.calcs_reversed:
            frequency_modes = calc.get("frequency_mode_vectors")
            ir_intensities = calc.get("IR_intens")
            ir_active = calc.get("IR_active")
            raman_intensities = calc.get("raman_intens")
            raman_active = calc.get("raman_active")

            # IR intensities and activities required
            # Raman intensitives/activities are optional
            if all(
                [x is not None for x in [frequency_modes, ir_intensities, ir_active]]
            ):
                break

        if frequency_modes is None:
            raise Exception("No frequency modes in task!")
        elif ir_intensities is None:
            raise Exception("No IR intensities in task!")
        elif ir_active is None:
            raise Exception("No IR activities in task!")

        ir_active = [True if ira.upper() == "YES" else False for ira in ir_active]
        if raman_active is not None:
            raman_active = [
                True if ra.upper() == "YES" else False for ra in raman_active
            ]

        warnings = list()
        if frequencies[0] < 0.0:
            warnings.append("Imaginary frequencies")

        id_string = f"vibrations-{molecule_id}-{task.task_id}-{task.lot_solvent}"
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            molecule=mol,
            frequencies=frequencies,
            frequency_modes=frequency_modes,
            ir_intensities=ir_intensities,
            ir_activities=ir_active,
            raman_intensities=raman_intensities,
            raman_activities=raman_active,
            warnings=warnings,
            origins=[MolPropertyOrigin(name="vibrations", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )



import pytest
from monty.serialization import loadfn
from emmet.core.molecules.summary import MoleculeSummaryDoc


@pytest.fixture(scope="session")
def docs_data(test_dir):
    raw = loadfn(test_dir / "3dfd058d8b422143d386d1ec9a723987-C1Li1O3-m1-1.json.gz")
    return raw


def test_summary_doc(docs_data):
    # Reformatting because summary build process changed
    actual_docs = dict()
    for k, v in docs_data.items():
        if k == "molecules":
            actual_docs[k] = v
        else:
            actual_docs[k] = v[0]

    summary_doc = MoleculeSummaryDoc.from_docs(
        molecule_id="3dfd058d8b422143d386d1ec9a723987-C1Li1O3-m1-1", docs=actual_docs
    )

    assert summary_doc.property_name == "summary"
    assert len(summary_doc.bonding) > 0
    assert len(summary_doc.thermo) > 0
    assert len(summary_doc.vibration) > 0
    assert len(summary_doc.redox) > 0
    assert len(summary_doc.orbitals) == 0
    assert len(summary_doc.metal_binding) == 0



from __future__ import annotations

from hashlib import blake2b
from typing import TYPE_CHECKING

from pydantic import BaseModel, Field

from emmet.core.molecules.metal_binding import MetalBindingData
from emmet.core.molecules.molecule_property import PropertyDoc
from emmet.core.mpid import MPID, MPculeID
from emmet.core.qchem.calc_types import CalcType, LevelOfTheory, TaskType
from emmet.core.types.enums import ValueEnum
from emmet.core.types.pymatgen_types.structure_adapter import MoleculeType
from emmet.core.utils import arrow_incompatible

if TYPE_CHECKING:
    from typing import Any

    from typing_extensions import Self

__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class HasProps(ValueEnum):
    """
    Enum of possible hasprops values.
    """

    molecules = "molecules"
    bonding = "bonding"
    metal_binding = "metal_binding"
    multipole_moments = "multipole_moments"
    orbitals = "orbitals"
    partial_charges = "partial_charges"
    partial_spins = "partial_spins"
    redox = "redox"
    thermo = "thermo"
    vibration = "vibration"


class ThermoComposite(BaseModel):
    """
    Summary information obtained from MoleculeThermoDocs
    """

    property_id: str | None = Field(
        None,
        description="Property ID map for this MoleculeThermoDoc",
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this MoleculeThermoDoc.",
    )

    electronic_energy: float | None = Field(
        None, description="Electronic energy of the molecule (units: eV)"
    )

    zero_point_energy: float | None = Field(
        None, description="Zero-point energy of the molecule (units: eV)"
    )

    total_enthalpy: float | None = Field(
        None, description="Total enthalpy of the molecule at 298.15K (units: eV)"
    )
    total_entropy: float | None = Field(
        None, description="Total entropy of the molecule at 298.15K (units: eV/K)"
    )

    free_energy: float | None = Field(
        None, description="Gibbs free energy of the molecule at 298.15K (units: eV)"
    )


class VibrationComposite(BaseModel):
    """
    Summary information obtained from VibrationDocs
    """

    property_id: str | None = Field(
        None,
        description="Property ID for this VibrationDoc.",
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this VibrationDoc.",
    )

    frequencies: list[float] | None = Field(
        None, description="List of molecular vibrational frequencies"
    )


class OrbitalComposite(BaseModel):
    """
    Summary information obtained from OrbitalDocs
    """

    property_id: str | None = Field(
        None,
        description="Property ID for this OrbitalDoc.",
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this OrbitalDoc.",
    )

    open_shell: bool | None = Field(
        None, description="Is this molecule open-shell (spin multiplicity != 1)?"
    )


class PartialChargesComposite(BaseModel):
    """
    Summary information obtained from PartialChargesDocs
    """

    property_id: str | None = Field(
        None,
        description="Property ID for this PartialChargesDoc.",
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this PartialChargesDoc.",
    )

    partial_charges: list[float] | None = Field(
        None,
        description="Atomic partial charges for the molecule",
    )


class PartialSpinsComposite(BaseModel):
    """
    Summary information obtained from PartialSpinsDocs
    """

    property_id: str | None = Field(
        None,
        description="Property ID for this PartialSpinsDoc.",
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this PartialSpinsDoc.",
    )

    partial_spins: list[float] | None = Field(
        None,
        description="Atomic partial spins for the molecule",
    )


class BondingComposite(BaseModel):
    """
    Summary information obtained from MoleculeBondingDocs
    """

    property_id: str | None = Field(
        None,
        description="Property ID for this MoleculeBondingDoc.",
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this MoleculeBondingDoc.",
    )

    bond_types: dict[str, list[float]] | None = Field(
        None,
        description="Dictionaries of bond types to their length, e.g. C-O to a list of the lengths of C-O bonds in "
        "Angstrom.",
    )

    bonds: list[tuple[int, int]] | None = Field(
        None,
        description="List of bonds. Each bond takes the form (a, b), where a and b are 0-indexed atom indices",
    )

    bonds_nometal: list[tuple[int, int]] | None = Field(
        None,
        description="List of bonds with all metal ions removed. Each bond takes the form in the form (a, b), where a "
        "and b are 0-indexed atom indices.",
    )


class MultipolesComposite(BaseModel):
    """
    Summary information obtained from ElectricMultipoleDocs
    """

    property_id: str | None = Field(
        None,
        description="Property ID for this ElectricMultipoleDoc.",
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this ElectricMultipoleDoc.",
    )

    total_dipole: float | None = Field(
        None,
        description="Total molecular dipole moment (Debye)",
    )

    resp_total_dipole: float | None = Field(
        None,
        description="Total dipole moment, calculated via restrained electrostatic potential (RESP) (Debye)",
    )


class RedoxComposite(BaseModel):
    """
    Summary information obtained from RedoxDocs
    """

    property_id: str | None = Field(None, description="Property ID for this RedoxDoc.")

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this RedoxDoc.",
    )

    electron_affinity: float | None = Field(
        None, description="Vertical electron affinity in eV"
    )

    ea_task_id: MPID | None = Field(
        None, description="Molecule ID for electron affinity"
    )

    ionization_energy: float | None = Field(
        None, description="Vertical ionization energy in eV"
    )

    ie_task_id: MPID | None = Field(
        None, description="Molecule ID for ionization energy"
    )

    reduction_free_energy: float | None = Field(
        None, description="Adiabatic free energy of reduction"
    )

    red_molecule_id: MPculeID | None = Field(
        None, description="Molecule ID for adiabatic reduction"
    )

    oxidation_free_energy: float | None = Field(
        None, description="Adiabatic free energy of oxidation"
    )

    ox_molecule_id: MPculeID | None = Field(
        None, description="Molecule ID for adiabatic oxidation"
    )

    reduction_potential: float | None = Field(
        None,
        description="Reduction potential referenced to the standard hydrogen electrode (SHE) (units: V)",
    )

    oxidation_potential: float | None = Field(
        None,
        description="Oxidation potential referenced to the standard hydrogen electrode (SHE) (units: V)",
    )


@arrow_incompatible
class MetalBindingComposite(BaseModel):
    """
    Summary information obtained from MetalBindingDocs
    """

    property_id: str | None = Field(
        None, description="Property ID for this MetalBindingDoc."
    )

    level_of_theory: str | None = Field(
        None,
        description="Level of theory for this MetalBindingDoc.",
    )

    binding_partial_charges_property_id: str | None = Field(
        None,
        description="ID of PartialChargesDoc used to estimate metal charge",
    )

    binding_partial_spins_property_id: str | None = Field(
        None,
        description="ID of PartialSpinsDoc used to estimate metal spin",
    )

    binding_partial_charges_lot_solvent: str | None = Field(
        None,
        description="Combination of level of theory and solvent used to calculate atomic partial charges",
    )

    binding_partial_spins_lot_solvent: str | None = Field(
        None,
        description="Combination of level of theory and solvent used to calculate atomic partial spins",
    )

    binding_charge_spin_method: str | None = Field(
        None,
        description="The method used for partial charges and spins (must be the same).",
    )

    binding_bonding_property_id: str | None = Field(
        None,
        description="ID of MoleculeBondingDoc used to detect bonding in this molecule",
    )

    binding_bonding_lot_solvent: str | None = Field(
        None,
        description="Combination of level of theory and solvent used to determine the coordination environment "
        "of the metal atom or ion",
    )

    binding_bonding_method: str | None = Field(
        None, description="The method used for to define bonding."
    )

    binding_thermo_property_id: str | None = Field(
        None,
        description="ID of MoleculeThermoDoc used to obtain this molecule's thermochemistry",
    )

    binding_thermo_lot_solvent: str | None = Field(
        None,
        description="Combination of level of theory and solvent used for uncorrected thermochemistry",
    )

    binding_thermo_correction_lot_solvent: str | None = Field(
        None,
        description="Combination of level of theory and solvent used to correct the electronic energy",
    )

    binding_thermo_combined_lot_solvent: str | None = Field(
        None,
        description="Combination of level of theory and solvent used for molecular thermochemistry, combining "
        "both the frequency calculation and (potentially) the single-point energy correction.",
    )

    binding_data: list[MetalBindingData] | None = Field(
        None, description="Binding data for each metal atom or ion in the molecule"
    )


@arrow_incompatible
class MoleculeSummaryDoc(PropertyDoc):
    """
    Summary information about molecules and their properties, useful for searching.
    """

    property_name: str = "summary"

    # molecules
    molecules: dict[str, MoleculeType] = Field(
        ...,
        description="The lowest energy optimized structures for this molecule for each solvent.",
    )

    molecule_levels_of_theory: dict[str, str] | None = Field(
        None,
        description="Level of theory used to optimize the best molecular structure for each solvent.",
    )

    species_hash: str | None = Field(
        None,
        description="Weisfeiler Lehman (WL) graph hash using the atom species as the graph "
        "node attribute.",
    )
    coord_hash: str | None = Field(
        None,
        description="Weisfeiler Lehman (WL) graph hash using the atom coordinates as the graph "
        "node attribute.",
    )

    inchi: str | None = Field(
        None, description="International Chemical Identifier (InChI) for this molecule"
    )
    inchi_key: str | None = Field(
        None, description="Standardized hash of the InChI for this molecule"
    )

    task_ids: list[MPID] = Field(
        [],
        title="Calculation IDs",
        description="List of Calculation IDs associated with this molecule.",
    )

    similar_molecules: list[MPculeID] = Field(
        [], description="IDs associated with similar molecules"
    )

    constituent_molecules: list[MPculeID] = Field(
        [],
        description="IDs of associated MoleculeDocs used to construct this molecule.",
    )

    unique_calc_types: list[CalcType] | None = Field(
        None,
        description="Collection of all unique calculation types used for this molecule",
    )

    unique_task_types: list[TaskType] | None = Field(
        None,
        description="Collection of all unique task types used for this molecule",
    )

    unique_levels_of_theory: list[LevelOfTheory] | None = Field(
        None,
        description="Collection of all unique levels of theory used for this molecule",
    )

    unique_solvents: list[str] | None = Field(
        None,
        description="Collection of all unique solvents (solvent parameters) used for this molecule",
    )

    unique_lot_solvents: list[str] | None = Field(
        None,
        description="Collection of all unique combinations of level of theory and solvent used for this molecule",
    )

    # Properties

    thermo: dict[str, ThermoComposite] | None = Field(
        None,
        description="A summary of thermodynamic data available for this molecule, organized by solvent",
    )

    vibration: dict[str, VibrationComposite] | None = Field(
        None,
        description="A summary of the vibrational data available for this molecule, organized by solvent",
    )

    orbitals: dict[str, OrbitalComposite] | None = Field(
        None,
        description="A summary of the orbital (NBO) data available for this molecule, organized by solvent",
    )

    partial_charges: dict[str, dict[str, PartialChargesComposite]] | None = Field(
        None,
        description="A summary of the partial charge data available for this molecule, organized by solvent and by "
        "method",
    )

    partial_spins: dict[str, dict[str, PartialSpinsComposite]] | None = Field(
        None,
        description="A summary of the partial spin data available for this molecule, organized by solvent and by "
        "method",
    )

    bonding: dict[str, dict[str, BondingComposite]] | None = Field(
        None,
        description="A summary of the bonding data available for this molecule, organized by solvent and by method",
    )

    multipole_moments: dict[str, MultipolesComposite] | None = Field(
        None,
        description="A summary of the electric multipole data available for this molecule, organized by solvent",
    )

    redox: dict[str, RedoxComposite] | None = Field(
        None,
        description="A summary of the redox data available for this molecule, organized by solvent",
    )

    metal_binding: dict[str, dict[str, MetalBindingComposite]] | None = Field(
        None,
        description="A summary of the metal binding data available for this molecule, organized by solvent and by "
        "method",
    )

    # has props
    has_props: dict[str, bool] | None = Field(
        None,
        description="Properties available for this molecule",
    )

    @classmethod
    def from_docs(cls, molecule_id: MPculeID, docs: dict[str, Any]) -> Self:
        """Converts a bunch of property docs into a SummaryDoc"""

        doc = _copy_from_docs(**docs)

        if len(doc["has_props"]) == 0:
            raise ValueError("Missing minimal properties!")

        id_string = f"summary-{molecule_id}"
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()
        doc["property_id"] = property_id

        return MoleculeSummaryDoc(molecule_id=molecule_id, **doc)


# Key mapping
summary_fields: dict[str, list] = {
    HasProps(k).value: v
    for k, v in {
        "molecules": [
            "charge",
            "spin_multiplicity",
            "natoms",
            "elements",
            "nelements",
            "composition",
            "composition_reduced",
            "formula_alphabetical",
            "chemsys",
            "symmetry",
            "molecules",
            "deprecated",
            "task_ids",
            "species_hash",
            "coord_hash",
            "inchi",
            "inchi_key",
            "unique_calc_types",
            "unique_task_types",
            "unique_levels_of_theory",
            "unique_solvents",
            "unique_lot_solvents",
            "similar_molecules",
            "constituent_molecules",
            "molecule_levels_of_theory",
        ],
        "thermo": [
            "electronic_energy",
            "zero_point_energy",
            "total_enthalpy",
            "total_entropy",
            "free_energy",
        ],
        "vibration": [
            "frequencies",
        ],
        "orbitals": [
            "open_shell",
        ],
        "partial_charges": ["partial_charges"],
        "partial_spins": ["partial_spins"],
        "bonding": ["bond_types", "bonds", "bonds_nometal"],
        "multipole_moments": [
            "total_dipole",
            "resp_total_dipole",
        ],
        "redox": [
            "electron_affinity",
            "ea_task_id",
            "ionization_energy",
            "ie_task_id",
            "reduction_free_energy",
            "red_molecule_id",
            "oxidation_free_energy",
            "ox_molecule_id",
            "reduction_potential",
            "oxidation_potential",
        ],
        "metal_binding": [
            "binding_partial_charges_property_id",
            "binding_partial_spins_property_id",
            "binding_partial_charges_lot_solvent",
            "binding_partial_spins_lot_solvent",
            "binding_charge_spin_method",
            "binding_bonding_property_id",
            "binding_bonding_lot_solvent",
            "binding_bonding_method",
            "binding_thermo_property_id",
            "binding_thermo_lot_solvent",
            "binding_thermo_correction_lot_solvent",
            "binding_thermo_combined_lot_solvent",
            "binding_data",
        ],
    }.items()
}


def _copy_from_docs(
    molecules: dict[str, Any],
    partial_charges: dict[str, dict[str, dict[str, Any]]] | None = None,
    partial_spins: dict[str, dict[str, dict[str, Any]]] | None = None,
    bonding: dict[str, dict[str, dict[str, Any]]] | None = None,
    metal_binding: dict[str, dict[str, dict[str, Any]]] | None = None,
    multipole_moments: dict[str, dict[str, Any]] | None = None,
    orbitals: dict[str, dict[str, Any]] | None = None,
    redox: dict[str, dict[str, Any]] | None = None,
    thermo: dict[str, dict[str, Any]] | None = None,
    vibration: dict[str, dict[str, Any]] | None = None,
):
    """Helper function to cut down documents to composite models and then combine to create a MoleculeSummaryDoc"""

    has_props: dict[str, bool] = {str(val.value): False for val in HasProps}  # type: ignore[attr-defined]
    d: dict[str, Any] = {"has_props": has_props, "origins": []}

    # Molecules is special because there should only ever be one
    # MoleculeDoc for a given molecule
    # There are not multiple MoleculeDocs for different solvents

    # NB: mypy misfires on enums. Using the following syntax works fine with mypy,
    # but HasProps.molecules.value can raise errors about attributes being defined
    d["has_props"][HasProps("molecules").value] = True
    for copy_key in summary_fields[HasProps("molecules").value]:
        d[copy_key] = molecules[copy_key]

    doc_type_mapping = {
        HasProps(k).value: v
        for k, v in {
            "partial_charges": (partial_charges, PartialChargesComposite),
            "partial_spins": (partial_spins, PartialSpinsComposite),
            "bonding": (bonding, BondingComposite),
            "metal_binding": (metal_binding, MetalBindingComposite),
            "multipole_moments": (multipole_moments, MultipolesComposite),
            "orbitals": (orbitals, OrbitalComposite),
            "redox": (redox, RedoxComposite),
            "thermo": (thermo, ThermoComposite),
            "vibration": (vibration, VibrationComposite),
        }.items()
    }

    by_method = {
        HasProps(k).value
        for k in {
            "partial_charges",
            "partial_spins",
            "bonding",
            "metal_binding",
        }
    }

    # Function to grab the keys and put them in the root doc
    for doc_key in summary_fields:
        sub_docs, target_type = doc_type_mapping.get(doc_key, (None, None))

        # No information for this particular set of properties
        # Shouldn't happen, but can
        if sub_docs is None or target_type is None:
            composite_docs = None
        else:
            composite_docs = dict()  # type: ignore[var-annotated]

            if isinstance(sub_docs, dict) and len(sub_docs) > 0:
                d["has_props"][doc_key] = True

                if doc_key in by_method:
                    for solvent, solv_entries in sub_docs.items():
                        composite_docs[solvent] = dict()
                        for method, entry in solv_entries.items():
                            composite_docs[solvent][method] = dict()
                            for copy_key in summary_fields[doc_key]:
                                composite_docs[solvent][method][copy_key] = entry.get(
                                    copy_key
                                )

                            composite_docs[solvent][method]["property_id"] = entry.get(
                                "property_id"
                            )
                            composite_docs[solvent][method]["level_of_theory"] = (
                                entry.get("level_of_theory")
                            )

                            # Convert to appropriate BaseModel
                            composite_docs[solvent][method] = target_type(
                                **composite_docs[solvent][method]
                            )

                else:
                    for solvent, entry in sub_docs.items():
                        composite_docs[solvent] = dict()
                        for copy_key in summary_fields[doc_key]:
                            composite_docs[solvent][copy_key] = entry.get(copy_key)

                        composite_docs[solvent]["property_id"] = entry.get(
                            "property_id"
                        )
                        composite_docs[solvent]["level_of_theory"] = entry.get(
                            "level_of_theory"
                        )

                        # Convert to appropriate BaseModel
                        composite_docs[solvent] = target_type(**composite_docs[solvent])

            d[doc_key] = composite_docs

    return d



from pydantic import BaseModel, Field

from emmet.core.types.enums import ValueEnum
from emmet.core.types.pymatgen_types.grain_boundary_adapter import GrainBoundaryType
from emmet.core.types.typing import DateTimeType


class GBTypeEnum(ValueEnum):
    """
    Grain boundary types
    """

    tilt = "tilt"
    twist = "twist"


class GrainBoundaryDoc(BaseModel):
    """
    Grain boundary energies, work of separation...
    """

    material_id: str | None = Field(
        None,
        description="The Materials Project ID of the material. This comes in the form: mp-******.",
    )

    sigma: int | None = Field(
        None,
        description="Sigma value of the boundary.",
    )

    type: GBTypeEnum | None = Field(
        None,
        description="Grain boundary type.",
    )

    rotation_axis: list[int] | None = Field(
        None,
        description="Rotation axis.",
    )

    gb_plane: list[int] | None = Field(
        None,
        description="Grain boundary plane.",
    )

    rotation_angle: float | None = Field(
        None,
        description="Rotation angle in degrees.",
    )

    gb_energy: float | None = Field(
        None,
        description="Grain boundary energy in J/m^2.",
    )

    initial_structure: GrainBoundaryType | None = Field(
        None, description="Initial grain boundary structure."
    )

    final_structure: GrainBoundaryType | None = Field(
        None, description="Final grain boundary structure."
    )

    pretty_formula: str | None = Field(
        None, description="Reduced formula of the material."
    )

    w_sep: float | None = Field(None, description="Work of separation in J/m^2.")

    cif: str | None = Field(None, description="CIF file of the structure.")

    chemsys: str | None = Field(
        None, description="Dash-delimited string of elements in the material."
    )

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent calculation for this Material document.",
    )



"""Define identifier schemas used in MP."""

from __future__ import annotations

from math import log, floor
import re
from pathlib import Path
from string import ascii_lowercase, digits
from typing import TYPE_CHECKING

from pydantic_core import CoreSchema, core_schema

# For dev_scripts compatibility, safe import this list
if (Path(__file__).parent / "_forbidden_alpha_id.py").exists():
    from emmet.core._forbidden_alpha_id import FORBIDDEN_ALPHA_ID_VALUES
else:
    FORBIDDEN_ALPHA_ID_VALUES = set()

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any
    from typing_extensions import Self

    from pydantic import GetJsonSchemaHandler
    from pydantic.json_schema import JsonSchemaValue

# matches "mp-1234" or "1234" followed by and optional "-(Alphanumeric)"
MPID_REGEX_PATTERN = r"^([A-Za-z]+-)?(\d+)(-[A-Za-z0-9]+)*$"
mpid_regex = re.compile(MPID_REGEX_PATTERN)

MPCULE_REGEX_PATTERN = (
    r"^([A-Za-z]+-)?([A-Fa-f0-9]+)-([A-Za-z0-9]+)-(m?[0-9]+)-([0-9]+)$"
)
mpculeid_regex = re.compile(MPCULE_REGEX_PATTERN)

# matches capital letters and numbers of length 26 (ULID)
# followed by and optional "-(Alphanumeric)"
check_ulid = re.compile(r"^[A-Z0-9]{26}(-[A-Za-z0-9]+)*$")

VALID_ALPHA_SEPARATORS: set[str] = {
    "-",
    ":",
    "_",
}


class MPID(str):
    """
    A Materials Project type ID with a prefix and an integer
    This class enables seemlessly mixing MPIDs and regular integer IDs
    Prefixed IDs are considered less than non-prefixed IDs to enable proper
    mixing with the Materials Project

    Args:
        val: Either 1) a prefixed string e.g. "mp-1234"
                    2) an integer e.g. 1234
                    3) a number stored as a string e.g. '1234'
                    4) an MPID

            Numbers stored as strings are coerced to ints
    """

    def __init__(self, val: MPID | int | str) -> None:
        if isinstance(val, MPID):
            self.parts = val.parts  # type: ignore
            self.string = val.string  # type: ignore

        elif isinstance(val, int) or (isinstance(val, str) and val.isnumeric()):
            if int(val) < 0:
                raise ValueError("MPID cannot represent a negative integer.")
            self.parts = ("", int(val))
            self.string = str(val)

        elif isinstance(val, str):
            if mpid_regex.fullmatch(val):
                parts = val.split("-")
                parts[1] = int(parts[1])  # type: ignore
                self.parts = tuple(parts)
            elif check_ulid.fullmatch(val):
                ulid = val.split("-")[0]
                self.parts = (ulid, 0)
            else:
                raise ValueError(
                    "MPID string representation must follow the format prefix-number or start with a valid ULID."
                )
            self.string = val

        else:
            raise ValueError(
                "Must provide an MPID, int, or string of the format prefix-number or start with a valid ULID."
            )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, MPID):
            return self.string == other.string
        elif isinstance(other, AlphaID):
            return other == self
        elif isinstance(other, (int, str)):
            return self.string == MPID(other).string
        return NotImplemented

    def __str__(self) -> str:
        return self.string

    def __repr__(self) -> str:
        return f"MPID({self})"

    def __lt__(self, other: MPID | int | str) -> bool:
        other_parts = MPID(other).parts

        if self.parts[0] and other_parts[0]:
            # both have prefixes; normal comparison
            return self.parts < other_parts
        elif self.parts[0]:
            # other is a pure int, self is prefixed
            # Always sort MPIDs before pure integer IDs
            return True
        elif other_parts[0]:
            # self is pure int, other is prefixed
            return False

        # both are pure ints; normal comparison
        return self.parts[1] < other_parts[1]

    def __gt__(self, other: MPID | int | str) -> bool:
        """Define greater than for MPID.

        Note that `__gt__` is not the same as `not __lt__`.
        If two values are equal, `__lt__` will return False.
        Defining
        ```__gt__ := not __lt__```
        will then incorrectly return True for equal values.
        """
        other_parts = MPID(other).parts

        if self.parts[0] and other_parts[0]:
            # both have prefixes; normal comparison
            return self.parts > other_parts
        elif not self.parts[0] and not other_parts[0]:
            # both are pure ints; normal comparison
            return self.parts[1] > other_parts[1]
        return not self.__lt__(other)

    def __hash__(self) -> int:
        return hash(self.string)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return dict(
            pattern=MPID_REGEX_PATTERN,
            examples=["mp-3534", "3453", "mp-834-Ag"],
            type="string",
        )

    @classmethod
    def validate(cls, __input_value: Any, _: core_schema.ValidationInfo):
        if isinstance(__input_value, MPID):
            return __input_value
        elif isinstance(__input_value, str) and mpid_regex.fullmatch(__input_value):
            return MPID(__input_value)
        elif isinstance(__input_value, str) and check_ulid.fullmatch(__input_value):
            return MPID(__input_value)
        elif isinstance(__input_value, int):
            return MPID(__input_value)

        raise ValueError("Invalid MPID Format")


class MPculeID(str):
    """
    A Materials Project Molecule ID with a prefix, hash, and two integer values
        representing the charge and spin of the molecule
    Unlike the MPID, you cannot compare raw hashes or raw integers to MPculeIDs

    Args:
        val: Either 1) an MPculeID
                    2) a prefixed string of format "prefix-hash-formula-charge-spin"
                    3) a non-prefixed string of format "hash-formula-charge-spin"

            Numbers stored as strings are coerced to ints
    """

    def __init__(self, val: MPculeID | str):
        if isinstance(val, MPculeID):
            self.parts = val.parts  # type: ignore
            self.string = val.string  # type: ignore

        elif isinstance(val, str):
            parts = val.split("-")
            if len(parts) == 4:
                parts[1] = int(parts[2].replace("m", "-"))  # type: ignore
                parts[2] = int(parts[3])  # type: ignore
            elif len(parts) == 5:
                parts[3] = int(parts[3].replace("m", "-"))  # type: ignore
                parts[4] = int(parts[4])  # type: ignore
            else:
                raise ValueError(
                    "MPculeID string representation must follow the "
                    "format prefix-hash-formula-charge-spin or hash-formula-charge-spin."
                )

            self.parts = tuple(parts)
            self.string = val

        else:
            raise ValueError(
                "Must provide an MPculeID, or string of the format prefix-hash-formula-charge-spin "
                "or hash-formula-charge-spin"
            )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, MPculeID):
            return self.string == other.string
        elif isinstance(other, str):
            return self.string == MPculeID(other).string
        return NotImplemented

    def __str__(self) -> str:
        return self.string

    def __repr__(self) -> str:
        return f"MPculeID({self})"

    def __lt__(self, other: MPculeID | str) -> bool:
        other_parts = MPculeID(other).parts

        return "-".join([str(x) for x in self.parts[-4:]]) < "-".join(
            [str(x) for x in other_parts[-4:]]
        )

    def __gt__(self, other: MPculeID | str) -> bool:
        return not self.__lt__(other)

    def __hash__(self) -> int:
        return hash(self.string)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return dict(
            pattern=r"^^([A-Za-z]+-)?([A-Fa-f0-9]+)-([A-Za-z0-9]+)-(m?[0-9]+)-([0-9]+)$",
            examples=[
                "1a525231bdac3f13e2fac0962fe8d053-Mg1-0-1",
                "22b40b99719ac570fc7e6225e855ec6e-F5Li1P1-m1-2",
                "mpcule-b9ba54febc77d2a9177accf4605767db-C1H41-2",
            ],
            type="string",
        )

    @classmethod
    def validate(cls, __input_value: Any, _: core_schema.ValidationInfo):
        if isinstance(__input_value, MPculeID):
            return __input_value
        elif isinstance(__input_value, str) and mpculeid_regex.fullmatch(__input_value):
            return MPculeID(__input_value)

        raise ValueError("Invalid MPculeID Format")


class AlphaID(str):
    """Identifier based on representing an integer as an alphabetical string.

    Args:
        _alphabet (str) : The alphabet to use, defaults to lowercase Roman.
        _identifier (str) : The alphabetical string identifier.
        _padlen (int) : The minimum length of an identifier to pad left with zeroes.
        _prefix (str) : The ID prefix, ex. "mp"
        _separator (str) : The separator between `_prefix` and `_identifier`, ex: "-"
        _cut_point (int or None) : For legacy purposes, all MPIDs minted before the
            transition to AlphaID will use the legacy format as `mp-<int>` when
            calling `AlphaID(...).string`.

            Thus `_cut_point`, if not `None`, defines the maximum MPID at which
            the integer in `AlphaID(...).string` is preferred.


    Notes:
        There is possibility for "obscene" strings to be used in AlphaID.
        For a comprehensive list of integer values of these IDs, you can
        import them:
        ```
        from pathlib import Path
        from importlib_resources import files as import_resource_file
        if (Path(import_resource_file("emmet.core")) / "_forbidden_alpha_id.py").exists():
            from emmet.core._forbidden_alpha_id import FORBIDDEN_ALPHA_ID_VALUES
        else:
            FORBIDDEN_ALPHA_ID_VALUES = set()
        ```

        Or run `generate_identifier_exclude_list.py` in `emmet-core/dev_scripts`
    """

    _identifier: str
    _padlen: int
    _prefix: str
    _separator: str
    _alphabet: str = ascii_lowercase
    _value: int | None = None
    _cut_point: int | None = 3347529

    def __new__(
        cls,
        identifier: str | int | MPID,
        padlen: int = 0,
        prefix: str | None = None,
        separator: str = "-",
    ) -> Self:
        """Define a new instance of AlphaID.

        Args:
            identifier (str, int, or MPID) : the identifier, either a string with characters belonging
                to AlphaID._alphabet, an integer to represent as a string, or an MPID.
            padlen (int, default = 0) : the amount of characters to pad to if a character
                string is too short. For example, the integer 149 has alpha representation "ft",
                and thus `AlphaID(149, padlen = 6)` would present as "aaaaft".
                Both "ft" and "aaaaft" have the same integer value.
            prefix (str or None, default = None) : if a str, the prefix to use.
                For example, `AlphaID(149, padlen = 6, prefix="mp")` would present as "mp-aaaaft".
            separator (str, default = "-") : the separator to use between the prefix and the
                string representation of the integer, if the prefix is non-empty.
                For example, `AlphaID(149, padlen = 6, prefix="mp", separator = ":")` would
                present as "mp:aaaaft".
        """

        if isinstance(identifier, str):
            if any(char.isupper() for char in identifier):
                raise ValueError(
                    "To avoid amibiguities, AlphaID does not permit uppercase letters."
                )
            non_alpha_char = set(identifier).difference(cls._alphabet + digits)
            valid_seps = non_alpha_char.intersection(VALID_ALPHA_SEPARATORS) > set()
            if len(non_alpha_char) == 1 and valid_seps:
                separator = list(non_alpha_char)[0]
                prefix, identifier = identifier.split(separator)

                if not prefix or not identifier:
                    raise ValueError(
                        "Missing prefix and/or identifer."
                        "To specify an AlphaID without a prefix, provide only the identifier."
                    )

            elif len(non_alpha_char) > 1 or (
                len(non_alpha_char) == 1 and not valid_seps
            ):
                raise ValueError(
                    f"Too many non-alpha-numeric characters: {', '.join(non_alpha_char)}"
                )

        elif isinstance(identifier, MPID):
            prefix, identifier = identifier.parts
            separator = "-"

        if isinstance(identifier, str) and set(identifier).intersection(digits):
            identifier = int(identifier)
        if isinstance(identifier, int):
            if identifier < 0:
                raise ValueError("AlphaID cannot represent a negative integer.")
            identifier = cls._integer_to_alpha_rep(identifier)

        prefix = prefix or ""
        if not prefix:
            separator = ""

        padded = max(0, padlen - len(identifier)) * cls._alphabet[0]
        if separator and separator not in VALID_ALPHA_SEPARATORS:
            raise ValueError(
                f"Invalid separator: {separator}. Use one of: {', '.join(VALID_ALPHA_SEPARATORS)}"
            )
        new_cls = str.__new__(cls, prefix + separator + padded + identifier)
        new_cls._identifier = identifier
        new_cls._padlen = padlen
        new_cls._prefix = prefix
        new_cls._separator = separator
        return new_cls

    @classmethod
    def _string_to_base_10_value(cls, string_value: str) -> int:
        """Obtain the integer value of an alphabetical string."""
        value = 0
        rev_codex = {letter: idx for idx, letter in enumerate(cls._alphabet)}
        base = len(cls._alphabet)
        for ipow, char in enumerate(string_value[::-1]):
            value += rev_codex[char] * base**ipow
        return value

    @classmethod
    def _integer_to_alpha_rep(cls, integer_val: int) -> str:
        """Obtain the string representation of an integer."""
        if integer_val == 0:
            return cls._alphabet[0]

        base = len(cls._alphabet)
        max_pow = floor(log(integer_val) / log(base))
        string: str = ""
        rem = integer_val
        for pow in range(max_pow, -1, -1):
            if rem == 0:
                string += cls._alphabet[0]
                continue
            mult = base**pow
            for coeff in range(base - 1, -1, -1):
                if coeff * mult <= rem:
                    string += cls._alphabet[coeff]
                    rem -= coeff * mult
                    break
            else:
                # Should never reach this point but still good to have this here
                raise ValueError(
                    f"Could not parse integer into string representation: {integer_val}"
                )
        return string

    def __hash__(self) -> int:
        """Ensure hashability."""
        return hash(str(self))

    def __int__(self) -> int:
        """Get and cache the current AlphaID's integer value."""
        if self._value is None:
            self._value = self._string_to_base_10_value(self._identifier)
        return self._value

    def __repr__(self) -> str:
        """Set AlphaID display name to distinguish from base string class."""
        return "AlphaID(" + self + ")"

    def __eq__(self, other: Any) -> bool:
        """Define equality of AlphaID.

        If other is an int, returns True if the integer value of the
        current instance equals the other.

        If other is a base str, returns True if the string representations
        are equal.

        If other is an AlphaID, returns True only if the prefix, separator,
        and value of the two are equal.
        """
        if isinstance(other, MPID) or (
            isinstance(other, str) and other.startswith("mp-")
        ):
            test = AlphaID(other)
        else:
            test = other

        if isinstance(test, int):
            return int(self) == test
        elif isinstance(test, AlphaID):
            return (
                test._prefix == self._prefix
                and test._separator == self._separator
                and int(self) == int(test)
            )
        elif isinstance(test, str):
            return str(self) == test
        raise NotImplementedError(f"Cannot compare AlphaID with {type(test)}")

    def __ne__(self, other: Any) -> bool:
        """Define inverse equality for AlphaID."""
        return not self.__eq__(other)

    @staticmethod
    def _coerce_value(
        alpha_id: AlphaID, other: Any, exception_on_mismatch: bool = True
    ) -> int:
        """Check if another value is comparable to a reference AlphaID.

        Used in defining `__add__`, `__sub__`, `__gt__`, and `__lt_``.

        Args:
            alpha_id : AlphaID
            other : Any other value
            exception_on_mismatch : bool (True = default)
                If True, raises an exception when the prefix and separators
                of two AlphaIDs do not match.

                To allow sorting of AlphaIDs, this is False for `__gt__` and `__lt__`
                but is True for `__add__` and `__sub__`.

        Returns:
            integer reprsenting the other value if possible.
            Raises exceptions if `other` cannot be compared to `alpha_id`.
        """
        test = AlphaID(other) if isinstance(other, MPID) else other

        if isinstance(test, AlphaID):
            if exception_on_mismatch:
                exc_str = ""
                if test._prefix != alpha_id._prefix:
                    exc_str += f"Prefixes do not match\n  left {alpha_id._prefix or None}\n  right {test._prefix or None}"

                if (
                    test._prefix
                    and alpha_id._prefix
                    and test._separator != alpha_id._separator
                ):
                    if exc_str:
                        exc_str += "\n"
                    exc_str += f"Separators do not match:\n  left {alpha_id._separator}\n  right {test._separator}"

                if exc_str:
                    raise TypeError(exc_str)

            diff = int(test)
        elif isinstance(test, str):
            diff = alpha_id._string_to_base_10_value(test)
        elif isinstance(test, int):
            diff = test
        else:
            raise NotImplementedError(f"Cannot compare AlphaID with type {type(test)}")

        return diff

    def __add__(self, other: Any) -> "AlphaID":
        """Define addition of AlphaID.

        Returns an AlphaID with the same `padlen` as the current instance, not `other`.
        Thus the order of addition can change the presentation of AlphaID, but not
        its value.

        If other is also an AlphaID, but its `prefix` and `separator` do not match,
        will not add the two. Only checks the separator if prefixes are both non-null.

        Args:
            other : the value to add to the current identifier.
                If a string, its integer value is first computed.
                The integer values are then added, and a new instance
                of AlphaID is returned.
        Returns:
            AlphaID representing the sum of the current and other values.
        """
        return AlphaID(
            int(self) + self._coerce_value(self, other),
            padlen=self._padlen,
            prefix=self._prefix,
            separator=self._separator,
        )

    def __sub__(self, other: Any) -> "AlphaID":
        """Define subtraction of AlphaID.

        See the docstring for `__add__`. The resultant `padlen` is taken
        from the current instance, not `other`.

        Will not subtract two AlphaIDs if `prefix` and `separator` do not match.
        """
        return self.__add__(-self._coerce_value(self, other))

    def __lt__(self, other: Any) -> bool:
        """Define AlphaID less than.

        Returns False between two AlphaIDs if `prefix` and `separator` do not match.
        """
        return int(self) < self._coerce_value(self, other, exception_on_mismatch=False)

    def __gt__(self, other: Any) -> bool:
        """Define AlphaID greater than.

        Returns False between two AlphaIDs if `prefix` and `separator` do not match.
        """
        return int(self) > self._coerce_value(self, other, exception_on_mismatch=False)

    @staticmethod
    def _format_legacy_ids(
        alpha_id: AlphaID,
        cutoff: int | None,
        prefix: str | None,
        separator: str | None,
        as_object: bool = False,
    ) -> str | MPID | AlphaID:
        """Format an AlphaID as either an MPID for legacy data, or AlphaID for newer data.

        Parameters
        -----------
        alpha_id : AlphaID
            An instance of AlphaID
        cutoff : int | None
            The largest legacy ID value to format as an MPID.
            If None, this returns the original AlphaID.
        prefix : str or None
            Identifier prefix
        separator : str or None
            Identifier separator. If returning an MPID and the prefix is not None,
            will override as a hyphen ("-")
        as_object : bool = False
            Whether to return objects (True) or a str (False)

        Returns
        -----------
        MPID if as_object and int(alpha_id) <= cutoff
        AlphaID if as_object and int(alpha_id) > cutoff
        str otherwise
        """
        if cutoff and ((v := int(alpha_id)) <= cutoff):
            idx = f"{prefix or ''}{'-' if prefix else ''}{v}"
            if as_object:
                return MPID(idx)
            return idx
        if as_object:
            return alpha_id
        return str(alpha_id)

    @property
    def string(self) -> str:
        """Legacy access to .string attr as in MPID."""
        return self._format_legacy_ids(
            self, self._cut_point, self._prefix, self._separator, as_object=False
        )

    @property
    def formatted(self) -> MPID | AlphaID:
        """Return an MPID for legacy indices, and AlphaID otherwise."""
        return self._format_legacy_ids(  # type: ignore[return-value]
            self, self._cut_point, self._prefix, self._separator, as_object=True
        )

    def copy(self) -> AlphaID:
        """Return a deep copy of the current instance."""
        return AlphaID(str(self))

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Any, handler: core_schema.CoreSchema
    ) -> core_schema.CoreSchema:
        """Generate pydantic schema for AlphaID."""
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        json_schema = handler(core_schema.str_schema())
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema["examples"] = ["mp-ft", "task:pqrs"]
        return json_schema

    @classmethod
    def validate(cls, __input_value: Any, _: core_schema.ValidationInfo) -> AlphaID:
        """Define pydantic validator for AlphaID."""
        if isinstance(__input_value, AlphaID):
            return __input_value
        elif isinstance(__input_value, str | int | MPID):
            return AlphaID(__input_value)

        raise ValueError(f"Invalid AlphaID Format {__input_value}")

    @property
    def parts(self) -> tuple[str, int]:
        """Mimic the parts attribute of MPID (prefix, integer value)."""
        return (
            self._prefix or "",
            int(self),
        )

    @property
    def next_safe(self) -> AlphaID:
        """Return the next multi-lingually "safe" AlphaID."""
        return _next_safe_alpha_id(self)


def _next_safe_alpha_id(
    start: int | str | AlphaID = 0,
) -> AlphaID:
    """Get the next "safe" (non-obscene) AlphaID.

    Returns the least AlphaID > AlphaID(start) such that its
    string value is "non-obscene".

    Uses a multi-lingual list of profanity from the shutter stock team
    (see emmet-core/dev_scripts/generate_identifier_exclude_list.py)
    to exclude generally-accepted crude language in AlphaIDs.

    Parameters
    -----------
    start : int | str | AlphaID = 0
        The AlphaID to start from.

    Returns
    -----------
    AlphaID : the next sequential AlphaID that is (mostly-)profanity free.
        (We're not perfect, there are so many languages.)
    """

    start_id = AlphaID(start) + 1
    while int(start_id) in FORBIDDEN_ALPHA_ID_VALUES:
        start_id += 1
    return start_id



"""Define enums with custom de-/serialization behavior.

Note that only the base classes for enums and enums
that are used widely across the code base should be
put here.

For example, `ValueEnum` is defined here, with custom
pydantic serialization features.

Similarly, `TaskState` is used across many sub-modules
in `emmet-core` and therefore is defined here.

Enums which only have one purpose / exist only within
one module, can and should remain in that module.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from pydantic_core import CoreSchema, core_schema

if TYPE_CHECKING:
    from typing import Any

    from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
    from typing_extensions import Self


class ValueEnum(Enum):
    """
    Enum that serializes to string as the value.

    While this method has an `as_dict` method, this
    returns a `str`. This is to ensure deserialization
    to a `str` when functions like `monty.json.jsanitize`
    are called on a ValueEnum with `strict = True` and
    `enum_values = False` (occurs often in jobflow).
    """

    def __str__(self):
        return str(self.value)

    def __eq__(self, obj: object) -> bool:
        """Special Equals to enable converting strings back to the enum"""
        if isinstance(obj, str):
            return super().__eq__(self.__class__(obj))
        elif isinstance(obj, self.__class__):
            return super().__eq__(obj)
        return False

    def __hash__(self):
        """Get a hash of the enum."""
        return hash(str(self))

    @classmethod
    def __get_pydantic_core_schema__(
        cls, _source_type: Any, _handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """Ensure pydantic (de)serialization."""

        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(),
                core_schema.with_info_plain_validator_function(cls.validate),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=from_str_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.value
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> dict[str, Any]:
        json_schema = handler(core_schema.str_schema())
        json_schema = handler.resolve_ref_schema(json_schema)
        return json_schema

    @classmethod
    def validate(cls, __input_value: Any, _: core_schema.ValidationInfo) -> Self:
        """Define pydantic validator for emmet enums."""
        if isinstance(__input_value, str):
            return (
                cls[__input_value]
                if __input_value in cls.__members__
                else cls(__input_value)
            )
        elif isinstance(__input_value, ValueEnum):
            return __input_value  # type: ignore[return-value]
        raise ValueError(f"Invalid {cls.__name__}: {__input_value}")


class DocEnum(ValueEnum):
    """
    Enum with docstrings support
    from: https://stackoverflow.com/a/50473952
    """

    def __new__(cls, value, doc=None):
        """add docstring to the member of Enum if exists

        Args:
            value: Enum member value
            doc: Enum member docstring, None if not exists
        """
        self = object.__new__(cls)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self


class IgnoreCaseEnum(ValueEnum):
    """Enum that permits case-insensitve lookup.

    Reference issue:
    https://github.com/materialsproject/api/issues/869
    """

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.upper() == value.upper():
                return member


class VaspObject(ValueEnum):
    """Types of VASP data objects."""

    BANDSTRUCTURE = "bandstructure"
    DOS = "dos"
    CHGCAR = "chgcar"
    AECCAR0 = "aeccar0"
    AECCAR1 = "aeccar1"
    AECCAR2 = "aeccar2"
    TRAJECTORY = "trajectory"
    ELFCAR = "elfcar"
    WAVECAR = "wavecar"
    LOCPOT = "locpot"
    OPTIC = "optic"
    PROCAR = "procar"


class StoreTrajectoryOption(ValueEnum):
    FULL = "full"
    PARTIAL = "partial"
    NO = "no"


class TaskState(ValueEnum):
    """
    VASP Calculation State
    """

    SUCCESS = "successful"
    FAILED = "failed"
    ERROR = "error"


class DeprecationMessage(DocEnum):
    MANUAL = "M", "Manual deprecation"
    SYMMETRY = (
        "S001",
        "Could not determine crystalline space group, needed for input set check.",
    )
    KPTS = "C001", "Too few KPoints"
    KSPACING = "C002", "KSpacing not high enough"
    ENCUT = "C002", "ENCUT too low"
    FORCES = "C003", "Forces too large"
    MAG = "C004", "At least one site magnetization is too large"
    POTCAR = (
        "C005",
        "At least one POTCAR used does not agree with the pymatgen input set",
    )
    CONVERGENCE = "E001", "Calculation did not converge"
    MAX_SCF = "E002", "Max SCF gradient too large"
    LDAU = "I001", "LDAU Parameters don't match the inputset"
    SET = ("I002", "Cannot validate due to missing or problematic input set")
    UNKNOWN = "U001", "Cannot validate due to unknown calc type"


class BatteryType(str, ValueEnum):
    """
    Enum for battery type
    """

    insertion = "insertion"
    conversion = "conversion"


class ThermoType(ValueEnum):
    """Thermodynamic hull types used in the database."""

    GGA_GGA_U = "GGA_GGA+U"
    GGA_GGA_U_R2SCAN = "GGA_GGA+U_R2SCAN"
    R2SCAN = "R2SCAN"
    UNKNOWN = "UNKNOWN"


class XasEdge(ValueEnum):
    """
    The interaction edge for XAS
    There are 2n-1 sub-components to each edge where
    K: n=1
    L: n=2
    M: n=3
    N: n=4
    """

    K = "K"
    L2 = "L2"
    L3 = "L3"
    L2_3 = "L2,3"


class XasType(ValueEnum):
    """
    The type of XAS Spectrum
    XANES - Just the near-edge region
    EXAFS - Just the extended region
    XAFS - Fully stitched XANES + EXAFS
    """

    XANES = "XANES"
    EXAFS = "EXAFS"
    XAFS = "XAFS"
